#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of a MLonMCU Run which represents a single benchmark instance for a given set of options."""
import itertools
import os
import copy
import tempfile
from pathlib import Path
from enum import IntEnum
from collections import defaultdict

from mlonmcu.logging import get_logger
from mlonmcu.artifact import ArtifactFormat, lookup_artifacts
from mlonmcu.config import str2bool
from mlonmcu.platform.platform import CompilePlatform, TargetPlatform, BuildPlatform, TunePlatform
from mlonmcu.report import Report  # TODO: move to mlonmcu.session.report
from mlonmcu.config import resolve_required_config, filter_config
from mlonmcu.feature.type import FeatureType
from mlonmcu.feature.features import get_matching_features, get_available_features
from mlonmcu.target.metrics import Metrics
from mlonmcu.models import SUPPORTED_FRONTENDS
from mlonmcu.models.model import Model, Program
from mlonmcu.platform import get_platforms
from mlonmcu.flow import SUPPORTED_FRAMEWORKS, SUPPORTED_BACKENDS

from .postprocess import SUPPORTED_POSTPROCESSES
from .postprocess.postprocess import RunPostprocess

logger = get_logger()


class RunStage(IntEnum):
    """Type describing the stages a run can have."""

    NOP = 0
    LOAD = 1  # unimplemented
    TUNE = 2
    BUILD = 3
    COMPILE = 4
    RUN = 5
    POSTPROCESS = 6
    DONE = 7


def add_any(new, base=None, append=True):
    ret = []
    if append:
        if isinstance(base, list):
            ret = base

    if isinstance(new, list):
        ret.extend(new)
    else:
        ret.append(new)
    return ret


class Run:
    """A run is single model/backend/framework/target combination with a given set of features and configs."""

    FEATURES = {"autotune", "target_optimized"}

    DEFAULTS = {
        "export_optional": False,
        "tune_enabled": False,
        "target_to_backend": True,
        "target_optimized_layouts": False,
        "target_optimized_schedules": False,
        "stage_subdirs": False,
    }

    REQUIRED = set()
    OPTIONAL = set()

    @classmethod
    def from_file(cls, path):
        """Restore a run object which was written to the disk."""
        raise NotImplementedError

    def __init__(
        self,
        idx=None,
        model=None,
        framework=None,
        frontends=None,
        backend=None,
        target=None,
        platforms=None,  # TODO: rename
        features=None,  # TODO: All features combined or explicit run-features -> postprocesses?
        config=None,  # TODO: All config combined or explicit run-config?
        postprocesses=None,
        archived=False,
        session=None,
        comment="",
    ):
        self.idx = idx
        self.model = model  # Model name, not object?
        self.frontends = frontends if frontends is not None else []
        self.framework = framework  # ???
        self.backend = backend
        self.platforms = platforms if platforms is not None else []
        self.artifacts_per_stage = {}
        self.archived = archived
        self.session = session
        self.postprocesses = postprocesses if postprocesses else []
        self.comment = comment
        # self.stage = RunStage.NOP  # max executed stage
        self.completed = {stage: stage == RunStage.NOP for stage in RunStage}

        self.directories = {}
        # self.init_directory()
        self.target = target
        self.cache_hints = []
        self.config = config if config else {}
        self.features = features if features else []
        self.run_config = {}
        self.run_features = self.process_features(features)
        self.run_config = filter_config(self.config, "run", self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        self.sub_names = []
        self.sub_parents = {}
        self.result = None
        self.failing = False  # -> RunStatus
        self.reason = None
        self.failed_stage = None
        # self.lock = threading.Lock()  # FIXME: use mutex instead of boolean
        self.locked = False
        self.report = None

    def process_features(self, features):
        """Utility which handles postprocess_features."""
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.RUN)
        for feature in features:
            assert feature.name in self.FEATURES, f"Incompatible feature: {feature.name}"
            feature.used = True
            tmp_run_config = {f"run.{key}": value for key, value in self.run_config.items()}
            feature.add_run_config(tmp_run_config)
            self.run_config = filter_config(tmp_run_config, "run", self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        return features

    @property
    def tune_enabled(self):
        """Get tune_enabled property."""
        value = self.run_config["tune_enabled"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def target_to_backend(self):
        """Get target_to_backend property."""
        value = self.run_config["target_to_backend"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def target_optimized_layouts(self):
        """Get target_optimized_layouts property."""
        value = self.run_config["target_optimized_layouts"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def target_optimized_schedules(self):
        """Get target_optimized_schedules property."""
        value = self.run_config["target_optimized_schedules"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def export_optional(self):
        """Get export_optional property."""
        value = self.run_config["export_optional"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def stage_subdirs(self):
        value = self.run_config["stage_subdirs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def build_platform(self):
        """Get platform for build stage."""
        if self.backend is not None and (
            hasattr(self.backend, "platform") and isinstance(self.backend.platform, BuildPlatform)
        ):
            return self.backend.platform
        for platform in self.platforms:
            if isinstance(platform, BuildPlatform):
                return platform
        return None

    @property
    def tune_platform(self):
        """Get platform for tune stage."""
        if self.backend is not None and (
            hasattr(self.backend, "platform") and isinstance(self.backend.platform, TunePlatform)
        ):
            return self.backend.platform
        for platform in self.platforms:
            if isinstance(platform, TunePlatform):
                return platform
        return None

    @property
    def compile_platform(self):
        """Get platform for compile stage."""
        if self.target is not None and isinstance(self.target.platform, CompilePlatform):
            return self.target.platform
        for platform in self.platforms:
            if isinstance(platform, CompilePlatform):
                return platform
        return None

    @property
    def target_platform(self):  # TODO: rename to run_stage?
        """Get platform for run stage."""
        if self.target:
            assert isinstance(self.target.platform, TargetPlatform)
            return self.target.platform
        for platform in self.platforms:
            if isinstance(platform, TargetPlatform):
                return platform
        return None

    def has_stage(self, stage):
        """Returns true if the given stage is available for this run."""
        if stage == RunStage.NOP:
            return True
        if stage == RunStage.LOAD:
            return self.model is not None and len(self.frontends) > 0
        if stage == RunStage.TUNE:
            return self.tune_enabled and self.backend is not None
        if stage == RunStage.BUILD:
            return self.backend is not None and self.framework is not None
        if stage == RunStage.COMPILE:
            return self.target is not None and len(self.platforms) > 0 and self.compile_platform is not None
        if stage == RunStage.RUN:
            return self.target is not None and len(self.platforms) > 0 and self.target_platform is not None
        if stage == RunStage.POSTPROCESS:
            return len(self.postprocesses) > 0
        if stage == RunStage.DONE:
            return False
        return False  # TODO: Throw error instead?

    @property
    def next_stage(self):
        """Determines the next not yet completed stage. Returns RunStage.DONE if already completed."""
        for stage in RunStage:
            if not self.completed[stage.value] and self.has_stage(stage):
                return stage
        return RunStage.DONE

    @property
    def last_stage(self):
        """Determines the next not yet completed stage. Returns RunStage.DONE if already completed."""
        last = None
        for stage in RunStage:
            if self.has_stage(stage):
                if not self.completed[stage.value]:
                    return last
                last = stage
        return None

    def lock(self):
        """Aquire a mutex to lock the current run."""
        # ret = self.lock.acquire(timeout=0)
        ret = not self.locked
        self.locked = True
        if not ret:
            raise RuntimeError("Parallel processing of the same run is not allowed")

    def unlock(self):
        """Release a mutex to unlock the current run."""
        # self.lock.release()
        self.locked = False

    def init_directory(self):
        """Initialize the temporary directory for this run."""
        if self.session is None:
            assert not self.archived
            self.tempdir = tempfile.TemporaryDirectory()
            self.dir = Path(self.tempdir.name)
        else:
            self.tempdir = None
            self.dir = self.session.runs_dir / str(self.idx)
            if not self.dir.is_dir():
                os.mkdir(self.dir)
            # This is not a good idea, but else we would need a mutex/lock on the shared build_dir
            # A solution would be to split up the framework runtime libs from the mlif...
            for platform in self.platforms:  # TODO: only do this if needed! (not for every platform)
                # The stage_subdirs setting is ignored here because platforms can be multi-stage!
                # platform.init_directory(path=Path(self.dir) / platform.name)
                if platform in self.directories:
                    continue
                platform_dir = Path(self.dir) / platform.name
                if platform.init_directory(path=platform_dir):
                    self.directories[platform.name] = platform_dir
            # if target not in self.directories:
            #     target_dir = Path(self.dir) /target.name
            #     if target.init_directory(path=target_dir)
            #         self.directories[target.name] = target_dir

            # TODO: other components

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "session":
                setattr(result, k, v)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def copy(self):
        """Create a new run based on this instance."""
        new = copy.deepcopy(self)
        if self.session:
            new_idx = self.session.request_run_idx()
            new.idx = new_idx
            # self.init_directory()
        return new

    def init_component(self, component_cls, context=None):
        """Helper function to create and configure a MLonMCU component instance for this run."""
        required_keys = component_cls.REQUIRED
        optional_keys = component_cls.OPTIONAL
        self.config.update(
            resolve_required_config(
                required_keys,
                optional=optional_keys,
                features=self.features,
                config=self.config,
                cache=context.cache if context else None,
                default_flags=context.environment.flags if context else None,
                hints=self.cache_hints,
            )
        )
        component_config = self.config.copy()  # TODOL get rid of this
        return component_cls(features=self.features, config=component_config)

    def add_model(self, model):
        """Setter for the model instance."""
        self.model = model
        self.model.config = filter_config(self.config, self.model.name, self.model.DEFAULTS, set(), set())
        for platform in self.platforms:
            self.model.add_platform_config(platform, platform.config)
            self.model.add_platform_defs(platform, platform.definitions)
        # TODO: update after load stage
        # for platform in self.platforms:
        #     self.add_platform_defs(platform.name, platform.definitions)

    def add_frontend(self, frontend, append=True):
        """Setter for the frontend instance."""
        self.frontends = add_any(frontend, self.frontends, append=append)
        for frontend in self.frontends:
            for platform in self.platforms:
                frontend.add_platform_config(platform, platform.config)
                frontend.add_platform_defs(platform.name, platform.definitions)

    def add_frontends(self, frontends, append=False):
        """Setter for the list of frontends."""
        self.frontends = add_any(frontends, self.frontends, append=append)
        for frontend in self.frontends:
            for platform in self.platforms:
                frontend.add_platform_config(platform, platform.config)
                frontend.add_platform_defs(platform.name, platform.definitions)

    def add_backend(self, backend):
        """Setter for the backend instance."""
        self.backend = backend
        # assert len(self.platforms) > 0, "Add at least a platform before adding a backend."
        if self.model is not None:
            if not isinstance(self.model, Model):
                self.backend = None
                return
            if not self.model.skip_check:
                assert self.backend.supports_model(self.model), (
                    "The added backend does not support the chosen model. "
                    "Add the backend before adding a model to find a suitable frontend."
                )
        for platform in self.platforms:
            self.backend.add_platform_config(platform.name, platform.config)
            self.backend.add_platform_defs(platform.name, platform.definitions)

    def add_framework(self, framework):
        """Setter for the framework instance."""
        self.framework = framework
        # assert len(self.platforms) > 0, "Add at least a platform before adding a framework."
        for platform in self.platforms:
            self.framework.add_platform_config(platform.name, platform.config)
            self.framework.add_platform_defs(platform.name, platform.definitions)

    def add_target(self, target):
        """Setter for the target instance."""
        self.target = target
        assert self.platforms is not None, "Add at least a platform before adding a target."
        for platform in self.platforms:
            self.target.add_platform_config(platform.name, platform.config)
            self.target.add_platform_defs(platform.name, platform.definitions)
        self.cache_hints = [self.target.get_arch()]
        # self.resolve_chache_refs()

    def add_platform(self, platform, append=True):
        """Setter for the platform instance."""
        self.platforms = add_any(platform, self.platforms, append=append)
        for frontend in self.frontends:
            self.frontend.add_platform_config(platform.name, platform.config)
            self.frontend.add_platform_defs(platform.name, platform.definitions)
        if self.model:
            self.model.add_platform_config(platform.name, platform.config)
            self.model.add_platform_defs(platform.name, platform.definitions)
        if self.backend:
            self.backend.add_platform_config(platform.name, platform.config)
            self.backend.add_platform_defs(platform.name, platform.definitions)
        if self.framework:
            self.framework.add_platform_config(platform.name, platform.config)
            self.framework.add_platform_defs(platform.name, platform.definitions)

    def add_platforms(self, platforms, append=False):
        """Setter for the list of platforms."""
        self.platforms = add_any(platforms, self.platforms, append=append)
        # TODO: check for duplicates?
        for platform in platforms:
            for frontend in self.frontends:
                self.frontend.add_platform_config(platform.name, platform.config)
                self.frontend.add_platform_defs(platform.name, platform.definitions)
            if self.model:
                self.model.add_platform_config(platform.name, platform.config)
                self.model.add_platform_defs(platform.name, platform.definitions)
            if self.backend:
                self.backend.add_platform_config(platform.name, platform.config)
                self.backend.add_platform_defs(platform.name, platform.definitions)
            if self.framework:
                self.framework.add_platform_config(platform.name, platform.config)
                self.framework.add_platform_defs(platform.name, platform.definitions)

    def add_postprocess(self, postprocess, append=True):
        """Setter for a postprocess instance."""
        self.postprocesses = add_any(postprocess, self.postprocesses, append=append)

    def add_postprocesses(self, postprocesses, append=False):
        """Setter for the list of postprocesses."""
        self.postprocesses = add_any(postprocesses, self.postprocesses, append=append)

    def add_feature(self, feature, append=True):
        """Setter for a feature instance."""
        self.features = add_any(feature, self.features, append=append)
        self.run_features = self.process_features(self.features)

    def add_features(self, features, append=False):
        """Setter for the list of features."""
        self.features = add_any(features, self.features, append=append)
        self.run_features = self.process_features(self.features)

    def add_model_by_name(self, model_name, context=None):
        """Helper function to initialize and configure a model by its name."""
        assert context is not None, "Please supply a context"
        assert len(self.frontends) > 0, "Add a frontend to the run before adding a model"
        model = None
        reasons = {}
        for frontend in self.frontends:
            if model is not None:
                break
            try:
                model_hints = frontend.lookup_models([model_name], config=self.config, context=context)
                # model_hints = lookup_models([model_name], frontends=self.frontends, context=context)
                for model_hint in model_hints:
                    if (
                        self.backend is None
                        or isinstance(model_hint, Program)
                        or (self.backend is not None and self.backend.supports_model(model_hint))
                    ):
                        self.frontends = [frontend]
                        assert model_hint is not None, "Unable to pick a suitable model"
                        model = model_hint
                        break
            except Exception as e:
                reasons[frontend.name] = str(e)
        if model is None:
            if reasons:
                logger.error("Lookup of model '%s' was not successfull. Reasons: %s", model_name, reasons)
            else:
                raise RuntimeError(f"Model with name '{model_name}' not found.")
        self.add_model(model)

    def add_frontend_by_name(self, frontend_name, context=None):
        """Helper function to initialize and configure a frontend by its name."""
        self.add_frontends_by_name([frontend_name], context=context)

    def add_frontends_by_name(self, frontend_names, context=None):
        """Helper function to initialize and configure frontends by their names."""
        frontends = []
        for name in frontend_names:
            try:
                assert context is not None and context.environment.has_frontend(
                    name
                ), f"The frontend '{name}' is not enabled for this environment"
                frontends.append(self.init_component(SUPPORTED_FRONTENDS[name], context=context))
            except Exception:
                continue
        assert len(frontends) > 0, "No compatible frontend was found"
        self.add_frontends(frontends)

    def add_backend_by_name(self, backend_name, context=None):
        """Helper function to initialize and configure a backend by its name."""
        assert context is not None and context.environment.has_backend(
            backend_name
        ), f"The backend '{backend_name}' is not enabled for this environment"
        if self.build_platform:
            self.add_backend(self.init_component(self.build_platform.create_backend(backend_name), context=context))
        else:
            self.add_backend(self.init_component(SUPPORTED_BACKENDS[backend_name], context=context))
        if self.backend is None:
            return
        framework_name = self.backend.framework  # TODO: does this work?
        assert context.environment.has_framework(
            framework_name
        ), f"The framework '{framework_name}' is not enabled for this environment"
        self.add_framework(self.init_component(SUPPORTED_FRAMEWORKS[framework_name], context=context))

    def add_target_by_name(self, target_name, context=None):
        """Helper function to initialize and configure a target by its name."""
        # We can not use the following code snipped are platform targets may be resolved dynamically
        # assert context is not None and context.environment.has_target(
        #     target_name
        # ), f"The target '{target_name}' is not enabled for this environment"
        assert len(self.platforms) > 0, "Please add a platform to the run before adding the target"
        self.add_target(self.init_component(self.target_platform.create_target(target_name), context=context))

    def add_platform_by_name(self, platform_name, context=None):
        """Helper function to initialize and configure a platform by its name."""
        self.add_platforms_by_name([platform_name], context=context)

    def add_platforms_by_name(self, platform_names, context=None):
        """Helper function to initialize and configure platforms by their names."""
        platforms = []
        for name in platform_names:
            assert context is not None and context.environment.has_platform(
                name
            ), f"The platform '{name}' is not enabled for this environment"
            platforms.append(self.init_component(get_platforms()[name], context=context))
        self.add_platforms(platforms)

    def add_postprocess_by_name(self, postprocess_name, append=True, context=None):
        """Helper function to initialize and configure a postprocesses by its name."""
        self.add_postprocesses_by_name([postprocess_name], append=append, context=context)

    def add_postprocesses_by_name(self, postprocess_names, append=False, context=None):
        """Helper function to initialize and configure postprocesses by their names."""
        postprocesses = []
        for name in postprocess_names:
            # TODO: ?
            # assert context is not None and context.environment.has_postprocess(
            #     postprocess_name
            # ), f"The postprocess '{postprocess_name}' is not enabled for this environment"
            postprocesses.append(self.init_component(SUPPORTED_POSTPROCESSES[name], context=context))
        self.add_postprocesses(postprocesses, append=append)

    def add_feature_by_name(self, feature_name, append=True, context=None):
        """Helper function to initialize and configure a feature by its name."""
        self.add_features_by_name([feature_name], context=context, append=append)

    def add_features_by_name(self, feature_names, append=False, context=None):
        """Helper function to initialize and configure features by their names."""
        features = []
        for feature_name in feature_names:
            available_features = get_available_features(feature_name=feature_name, deps=True)
            # check for already added features
            temp = self.features + features if append else features
            added_names = [f.name for f in temp]
            for feature_name_, feature_cls_ in available_features.items():
                if feature_name_ in added_names:
                    continue
                features.append(self.init_component(feature_cls_, context=context))
        self.add_features(features, append=append)

    def __repr__(self):
        probs = []
        if self.model:
            probs.append(str(self.model))
        if len(self.platforms) > 0:
            probs.append(str(self.platforms[0] if len(self.platforms) == 1 else self.platforms))
        if len(self.frontends) > 0:
            probs.append(str(self.frontends[0] if len(self.frontends) == 1 else self.frontends))
        if self.backend:
            probs.append(str(self.backend))
        if self.target:
            probs.append(str(self.target))
        if self.features and len(self.features) > 0:
            probs.append(str(self.features))
        if self.config and len(self.config) > 0:
            probs.append(str(self.config))
        return "Run(" + ",".join(probs) + ")"

    @property
    def frontend(self):
        assert len(self.frontends) > 0, "Not frontend is available for this run."
        return self.frontends[0]

    @property
    def artifacts(self):
        sub = "default"
        ret = sum(list(itertools.chain([subs[sub] for stage, subs in self.artifacts_per_stage.items()])), [])
        return ret

    def get_all_sub_artifacts(self, sub, stage=None):
        if sub is None:
            return []
        assert sub in self.sub_names
        if stage is None:
            stage = self.last_stage
        parents = self.sub_parents[(stage, sub)]
        stage_, sub_ = parents
        artifacts = self.get_all_sub_artifacts(sub_, stage_) + self.artifacts_per_stage[stage][sub]
        return artifacts

    def toDict(self):
        """Utility not implemented yet. (TODO: remove?)"""
        raise NotImplementedError

    def export_stage(self, stage, optional=False):
        """Export stage artifacts of this run to its directory."""
        # TODO: per stage subdirs?
        subdir = self.stage_subdirs
        if stage in self.artifacts_per_stage:
            for name in self.artifacts_per_stage[stage]:
                artifacts = self.artifacts_per_stage[stage][name]
                for artifact in artifacts:
                    if not artifact.optional or optional:
                        dest = self.dir
                        if subdir:
                            stage_idx = int(stage)
                            dest = dest / "stages" / str(stage_idx)
                            # TODO: stages.txt for mapping between stage idx and name
                        if name not in ["", "default"]:
                            dest = dest / "sub" / name
                        dest.mkdir(parents=True, exist_ok=True)
                        extract = artifact.fmt in [ArtifactFormat.MLF, ArtifactFormat.ARCHIVE]
                        # extract = artifact.fmt == ArtifactFormat.MLF
                        # and not isinstance(self.platform, MicroTvmPlatform)
                        artifact.export(dest)
                        # Keep the tar as well as the extracted files
                        if extract:
                            artifact.export(dest, extract=True)

    def postprocess(self):
        """Postprocess the 'run'."""
        logger.debug("%s Processing stage POSTPROCESS", self.prefix)
        self.lock()
        # assert self.completed[RunStage.RUN]  # Alternative: allow to trigger previous stages recursively as a fallback

        def _merge_dicts_of_lists(*args):
            if len(args) == 0:
                return {}
            a = args[0]
            de = defaultdict(list, a)
            for b in args[1:]:
                for (
                    i,
                    j,
                ) in b.items():
                    de[i].extend(j)
            return dict(de)

        self.artifacts_per_stage[RunStage.POSTPROCESS] = {"default": []}
        temp_report = self.get_report()
        last_stage = self.last_stage
        assert last_stage is not None
        self.export_stage(last_stage, optional=self.export_optional)
        for name in self.artifacts_per_stage[last_stage]:
            merged = {"default": []}
            before = self.get_all_sub_artifacts(name)
            for postprocess in self.postprocesses:
                if isinstance(postprocess, RunPostprocess):
                    artifacts = postprocess.post_run(temp_report, before)
                    before.extend(artifacts)
                    if artifacts is None:
                        artifacts = []
                    new = {}
                    if isinstance(artifacts, dict):
                        new.update(
                            {
                                (
                                    key
                                    if name in ["", "default"]
                                    else (f"{name}_{key}" if key not in ["", "default"] else name)
                                ): value
                                for key, value in artifacts.items()
                            }
                        )
                    else:
                        new.update({name if name in ["", "default"] else f"{name}": artifacts})
                    merged = _merge_dicts_of_lists(merged, new)
            self.artifacts_per_stage[RunStage.POSTPROCESS].update(merged)
            self.sub_parents.update({(RunStage.POSTPROCESS, key): (self.last_stage, name) for key in merged.keys()})
        self.sub_names.extend(self.artifacts_per_stage[RunStage.POSTPROCESS])
        self.sub_names = list(set(self.sub_names))

        self.report = temp_report

        self.completed[RunStage.POSTPROCESS] = True
        self.unlock()
        self.export_stage(RunStage.POSTPROCESS, optional=self.export_optional)

    def run(self):
        """Run the 'run' using the defined target."""
        logger.debug("%s Processing stage RUN", self.prefix)
        self.lock()
        # Alternative: drop artifacts of higher stages when re-triggering a lower one?
        self.artifacts_per_stage[RunStage.RUN] = {}
        if self.has_stage(RunStage.COMPILE):
            assert self.completed[RunStage.COMPILE]
            self.export_stage(RunStage.COMPILE, optional=self.export_optional)
            for name in self.artifacts_per_stage[RunStage.COMPILE]:
                elf_artifact = self.artifacts_per_stage[RunStage.COMPILE][name][0]
                artifacts = self.target.generate_artifacts(elf_artifact.path)
                if isinstance(artifacts, dict):
                    new = {
                        (
                            key
                            if name in ["", "default"]
                            else (f"{name}_{key}" if key not in ["", "default"] else name)
                        ): value
                        for key, value in artifacts.items()
                    }
                else:
                    new = {name if name in ["", "default"] else f"{name}": artifacts}
                self.artifacts_per_stage[RunStage.RUN].update(new)
                self.sub_parents.update({(RunStage.RUN, key): (self.last_stage, name) for key in new.keys()})
        else:
            assert self.completed[RunStage.BUILD]  # Used for tvm platform
            self.export_stage(RunStage.BUILD, optional=self.export_optional)
            for name in self.artifacts_per_stage[RunStage.BUILD]:
                shared_object_artifact = self.artifacts_per_stage[RunStage.BUILD][name][0]
                artifacts = self.target.generate_artifacts(shared_object_artifact.path)
                if isinstance(artifacts, dict):
                    new = {
                        (
                            key
                            if name in ["", "default"]
                            else (f"{name}_{key}" if key not in ["", "default"] else name)
                        ): value
                        for key, value in artifacts.items()
                    }
                else:
                    new = {name if name in ["", "default"] else f"{name}": artifacts}
                self.artifacts_per_stage[RunStage.RUN].update(new)
                self.sub_parents.update({(RunStage.RUN, key): (self.last_stage, name) for key in new.keys()})
        self.sub_names.extend(self.artifacts_per_stage[RunStage.RUN])
        self.sub_names = list(set(self.sub_names))

        self.completed[RunStage.RUN] = True
        self.unlock()

    def compile(self):
        """Compile the target software for the run."""
        logger.debug("%s Processing stage COMPILE", self.prefix)
        self.lock()
        if isinstance(self.model, Model):
            assert self.completed[RunStage.BUILD]

            self.export_stage(RunStage.BUILD, optional=self.export_optional)
            self.artifacts_per_stage[RunStage.COMPILE] = {}
            for name in self.artifacts_per_stage[RunStage.BUILD]:
                codegen_dir = self.dir if not self.stage_subdirs else (self.dir / "stages" / str(int(RunStage.BUILD)))
                if name not in ["", "default"]:
                    codegen_dir = codegen_dir / "sub" / name
                # TODO!
                artifacts = self.compile_platform.generate_artifacts(
                    codegen_dir, self.target
                )  # TODO: has to go into different dirs
                # artifacts = self.compile_platform.artifacts
                if isinstance(artifacts, dict):
                    new = {
                        (
                            key
                            if name in ["", "default"]
                            else (f"{name}_{key}" if key not in ["", "default"] else name)
                        ): value
                        for key, value in artifacts.items()
                    }
                else:
                    new = {name if name in ["", "default"] else f"{name}": artifacts}
                self.artifacts_per_stage[RunStage.COMPILE].update(new)
                self.sub_parents.update({(RunStage.COMPILE, key): (self.last_stage, name) for key in new.keys()})
        else:
            assert self.completed[RunStage.LOAD]
            self.artifacts_per_stage[RunStage.COMPILE] = {}
            codegen_dir = self.dir if not self.stage_subdirs else (self.dir / "stages" / str(int(RunStage.BUILD)))
            name = "default"
            artifacts = self.compile_platform.generate_artifacts(codegen_dir, self.target)
            if isinstance(artifacts, dict):
                new = {
                    key if name in ["", "default"] else (f"{name}_{key}" if key not in ["", "default"] else name): value
                    for key, value in artifacts.items()
                }
            else:
                new = {name if name in ["", "default"] else f"{name}": artifacts}
            self.artifacts_per_stage[RunStage.COMPILE].update(new)
            self.sub_parents.update({(RunStage.COMPILE, key): (self.last_stage, name) for key in new.keys()})
        self.sub_names.extend(self.artifacts_per_stage[RunStage.COMPILE])
        self.sub_names = list(set(self.sub_names))

        self.completed[RunStage.COMPILE] = True
        self.unlock()

    def build(self):
        """Process the run using the choosen backend."""
        logger.debug("%s Processing stage BUILD", self.prefix)
        self.lock()
        assert (not self.has_stage(RunStage.TUNE)) or self.completed[RunStage.TUNE]

        target_to_backend = self.target_to_backend and (self.target is not None)
        if self.backend.needs_target or self.target_optimized_layouts or self.target_optimized_schedules:
            assert self.target is not None, "Config target_to_backend can only be used if a target was provided"
            target_to_backend = True

        if target_to_backend:
            self.target.add_backend_config(
                self.backend.name,
                self.backend.config,
                optimized_layouts=self.target_optimized_layouts,
                optimized_schedules=self.target_optimized_schedules,
            )

        def _build():
            # TODO: allow raw data as well as filepath in backends
            artifacts = self.backend.generate_artifacts()
            if isinstance(artifacts, dict):
                new = {
                    key if name in ["", "default"] else (f"{name}_{key}" if key not in ["", "default"] else name): value
                    for key, value in artifacts.items()
                }
            else:
                new = {name if name in ["", "default"] else f"{name}": artifacts}
            self.artifacts_per_stage[RunStage.BUILD].update(new)
            self.sub_parents.update({(RunStage.BUILD, key): (self.last_stage, name) for key in new.keys()})

        self.artifacts_per_stage[RunStage.BUILD] = {}
        if self.has_stage(RunStage.TUNE):
            self.export_stage(RunStage.TUNE, optional=self.export_optional)
            for name in self.artifacts_per_stage[RunStage.TUNE]:
                tune_stage_artifacts = self.artifacts_per_stage[RunStage.TUNE][name]
                tuning_artifact = lookup_artifacts(
                    tune_stage_artifacts, fmt=ArtifactFormat.TEXT, flags=["records"], first_only=True
                )
                if len(tuning_artifact) == 0:
                    # fallback for metascheduler
                    tuning_artifact = lookup_artifacts(
                        tune_stage_artifacts,
                        fmt=ArtifactFormat.ARCHIVE,
                        flags=["records", "metascheduler"],
                        first_only=True,
                    )
                    if len(tuning_artifact) == 0:
                        continue
                tuning_artifact = tuning_artifact[0]
                if not tuning_artifact.exported:
                    tuning_artifact.export(self.dir)
                tuner_name = tuning_artifact.flags[-1]  # TODO: improve
                self.backend.set_tuning_records(tuning_artifact.path, tuner_name=tuner_name)
                candidate = (RunStage.TUNE, name)
                assert candidate in self.sub_parents
                parent_stage, parent_name = self.sub_parents[candidate]
                assert parent_stage == RunStage.LOAD
                assert parent_name in self.artifacts_per_stage[RunStage.LOAD]
                load_stage_artifacts = self.artifacts_per_stage[parent_stage][parent_name]
                model_artifact = lookup_artifacts(load_stage_artifacts, flags=["model"], first_only=True)
                assert len(model_artifact) > 0
                model_artifact = model_artifact[0]
                if not model_artifact.exported:
                    model_artifact.export(self.dir)
                input_shapes = self.model.input_shapes
                output_shapes = self.model.output_shapes
                input_types = self.model.input_types
                output_types = self.model.output_types
                self.backend.load_model(
                    model=model_artifact.path,
                    input_shapes=input_shapes,
                    output_shapes=output_shapes,
                    input_types=input_types,
                    output_types=output_types,
                )
                _build()

        else:
            self.export_stage(RunStage.LOAD, optional=self.export_optional)  # Not required anymore?
            for name in self.artifacts_per_stage[RunStage.LOAD]:
                load_stage_artifacts = self.artifacts_per_stage[RunStage.LOAD][name]
                model_artifact = lookup_artifacts(load_stage_artifacts, flags=["model"], first_only=True)
                if len(model_artifact) == 0:
                    # TODO: This breaks because number of subs can not decrease...
                    continue
                assert len(model_artifact) == 1
                model_artifact = model_artifact[0]
                if not model_artifact.exported:
                    model_artifact.export(self.dir)
                input_shapes = self.model.input_shapes
                output_shapes = self.model.output_shapes
                input_types = self.model.input_types
                output_types = self.model.output_types
                self.backend.load_model(
                    model=model_artifact.path,
                    input_shapes=input_shapes,
                    output_shapes=output_shapes,
                    input_types=input_types,
                    output_types=output_types,
                )
                _build()

        self.sub_names.extend(self.artifacts_per_stage[RunStage.BUILD])
        self.sub_names = list(set(self.sub_names))

        self.completed[RunStage.BUILD] = True
        self.unlock()

    def tune(self):
        """Tune the run using the choosen backend (if supported)."""
        logger.debug("%s Processing stage TUNE", self.prefix)
        self.lock()
        assert self.completed[RunStage.LOAD]

        target_to_backend = self.target_to_backend and (self.target is not None)
        if self.backend.needs_target or self.target_optimized_layouts or self.target_optimized_schedules:
            assert self.target is not None, "Backend needs target"
            target_to_backend = True

        if target_to_backend:
            self.target.add_backend_config(
                self.backend.name,
                self.backend.config,
                optimized_layouts=self.target_optimized_layouts,
                optimized_schedules=self.target_optimized_schedules,
            )

        self.export_stage(RunStage.LOAD, optional=self.export_optional)
        self.artifacts_per_stage[RunStage.TUNE] = {}
        for name in self.artifacts_per_stage[RunStage.LOAD]:
            model_artifact = self.artifacts_per_stage[RunStage.LOAD][name][0]
            # if not model_artifact.exported:
            #     model_artifact.export(self.dir)

            # TODO: allow raw data as well as filepath in backends
            assert self.tune_platform, "Autotuning requires a TunePlatform"
            res = self.tune_platform.tune_model(model_artifact.path, self.backend, self.target)
            new = {f"{name}": []}
            if res:
                if isinstance(res, dict):
                    new = {key if name in ["", "default"] else f"{name}_{key}": value for key, value in res.items()}
                else:
                    new = {name if name in ["", "default"] else f"{name}": res}
            self.artifacts_per_stage[RunStage.TUNE].update(new)
            self.sub_parents.update({(RunStage.TUNE, key): (RunStage.LOAD, name) for key in new.keys()})
        self.sub_names.extend(self.artifacts_per_stage[RunStage.TUNE])
        self.sub_names = list(set(self.sub_names))

        self.completed[RunStage.TUNE] = True
        self.unlock()

    def load(self):
        """Load the model using the given frontend."""
        logger.debug("%s Processing stage LOAD", self.prefix)
        self.lock()
        # assert self.completed[RunStage.NOP]

        artifacts = self.frontend.generate_artifacts(self.model)
        # The following is very very dirty but required to update arena sizes via model metadata...
        cfg_new = {}
        if isinstance(self.model, Model):
            self.frontend.process_metadata(self.model, cfg=cfg_new)
            if len(cfg_new) > 0:
                for key, value in cfg_new.items():
                    component, name = key.split(".")[:2]
                    if self.backend is not None and component == self.backend.name:
                        self.backend.config[name] = value
                    elif component == self.model.name:
                        # Do not overwrite user-provided shapes and types
                        if self.model.config[name] is None:
                            # self.model.config[name] = value
                            self.model.config = filter_config(
                                {key: value}, self.model.name, self.model.config, set(), set()
                            )
                    else:
                        for platform in self.platforms:
                            if platform is not None and component == platform.name:
                                platform.config[name] = value
                    self.config[key] = value
        if isinstance(artifacts, dict):
            self.artifacts_per_stage[RunStage.LOAD] = artifacts
        else:
            self.artifacts_per_stage[RunStage.LOAD] = {"default": artifacts}
        self.sub_names.extend(self.artifacts_per_stage[RunStage.LOAD])
        self.sub_names = list(set(self.sub_names))
        self.sub_parents.update(
            {(RunStage.LOAD, key): (None, None) for key in self.artifacts_per_stage[RunStage.LOAD].keys()}
        )

        self.completed[RunStage.LOAD] = True
        self.unlock()

    def process(self, until=RunStage.RUN, skip=None, export=False):
        """Process the run until a given stage."""
        skip = skip if skip is not None else []
        if until == RunStage.DONE:
            until = RunStage.DONE - 1
        start = self.next_stage  # self.stage hold the max finished stage
        if until < start:
            logger.debug("%s Nothing to do", self.prefix)
            return self.get_report()

        if start > RunStage.NOP:
            logger.debug(
                # self.prefix + "Processing run until stage %s: %s",
                "%s Continuing run from stage %s until stage %s",
                self.prefix,
                str(RunStage(start).name),
                str(RunStage(until).name),
                # str(self),
            )
        else:
            logger.debug(
                # self.prefix + "Processing run until stage %s: %s",
                "%s Processing run until stage %s",
                self.prefix,
                str(RunStage(until).name),
                # str(self),
            )
        for stage in range(start, until + 1):
            if not self.has_stage(stage):
                continue
            if stage in skip:
                logger.debug("%s Skipping stage %s", self.prefix, str(RunStage(stage).name))
                continue
            if stage < RunStage.POSTPROCESS:
                self.report = None  # Regenerate report if earlier stages are executed
            stage_funcs = {
                RunStage.NOP: lambda *args, **kwargs: None,  # stage already done
                RunStage.LOAD: self.load,
                RunStage.TUNE: self.tune,
                RunStage.BUILD: self.build,
                RunStage.COMPILE: self.compile,
                RunStage.RUN: self.run,
                RunStage.POSTPROCESS: self.postprocess,
            }
            func = stage_funcs[stage]
            if func:
                self.failing = False
                try:
                    func()
                except Exception as e:
                    self.failing = True
                    self.reason = e
                    if self.locked:
                        self.unlock()
                    logger.exception(e)
                    run_stage = RunStage(stage).name
                    self.failed_stage = run_stage
                    logger.error("%s Run failed at stage '%s', aborting...", self.prefix, run_stage)
                    break
            # self.stage = stage  # FIXME: The stage_func should update the stage intead?
        report = self.get_report()
        if export:
            self.export(optional=self.export_optional)  # TODO: set to flase?
            report_file = Path(self.dir) / "report.csv"
            report.export(report_file)
        return report

    def write_run_file(self):
        """Create a run.txt file which contains information used to reconstruct the run based
        on its properties at a later point in time."""
        logger.debug("%s Writing run file", self.prefix)
        filename = self.dir / "run.txt"
        with open(filename, "w", encoding="utf-8") as handle:
            handle.write(str(self))

    @property
    def prefix(self):
        """Get prefix property."""
        return (
            (f"[session-{self.session.idx}] [run-{self.idx}]" if self.session else f"[run-{self.idx}]")
            if self.idx is not None
            else ""
        )

    def get_all_feature_names(self, only_used=True):
        """Return list of feature names for this run."""
        return [feature.name for feature in self.features if feature.used or not only_used]

    def get_all_postprocess_names(self):
        """Return list of postprocess names for this run."""
        return [postprocess.name for postprocess in self.postprocesses]

    def get_frontend_name(self):
        """Return frontend name(s) for this run."""
        ret = [frontend.name for frontend in self.frontends]
        return ret[0] if len(ret) == 1 else ret

    def get_platform_name(self):
        """Return platform name(s) for this run."""
        used = list(set([self.tune_platform, self.build_platform, self.compile_platform, self.target_platform]))
        ret = [platform.name for platform in used if platform is not None]
        return ret[0] if len(ret) == 1 else ret

    def get_reason_text(self):
        ret = str(type(self.reason).__name__) if self.reason else None
        if self.failed_stage:
            ret += " @ " + str(self.failed_stage)
        return ret

    def get_all_configs(self, omit_paths=False, omit_defaults=False, omit_globals=False):
        """Return dict with component-specific and global configuration for this run."""

        def has_prefix(key):
            """Returns true if the configuration key does not have global scope."""
            return "." in key

        def config_helper(obj, prefix=None):
            """Helper to access the configuration of a given component object."""
            if prefix:
                name = prefix
            else:
                assert hasattr(obj, "name")
                name = obj.name
            omit_list = []
            defaults = obj.DEFAULTS
            for key, value in obj.config.items():
                if omit_defaults and key in defaults and defaults[key] == value:
                    omit_list.append(key)
            ret = {
                key if has_prefix(key) else f"{name}.{key}": value
                for key, value in obj.config.items()
                if key not in omit_list
            }
            if omit_paths:
                ret = {
                    key: value
                    for key, value in ret.items()
                    if not (
                        isinstance(value, Path)
                        or (isinstance(value, str) and len(str(value)) < 200 and Path(value).exists())
                    )
                }
            return ret

        ret = {}
        if not omit_globals:
            ret.update(
                {key: value for key, value in self.config.items() if not has_prefix(key)}
            )  # Only config without a prefix!
        ret.update(config_helper(self.model))
        for frontend in self.frontends:
            ret.update(config_helper(frontend))
        if self.backend:
            ret.update(config_helper(self.backend))
        if self.framework:
            ret.update(config_helper(self.framework))
        if self.target:
            self.target.reconfigure()
            ret.update(config_helper(self.target))
        for platform in self.platforms:
            ret.update(config_helper(platform))
        for feature in self.features:
            ret.update(config_helper(feature))
        for postprocess in self.postprocesses:
            ret.update(config_helper(postprocess))
        return ret

    def get_report(self):
        """Returns teh complete report of this run."""
        if self.completed[RunStage.POSTPROCESS]:
            if self.report is not None:
                return (
                    self.report
                )  # Use postprocessed report instead of generating a new one (TODO: find a better approach)
        # TODO: config or args for stuff like (session id) and run id as well as detailed features and configs
        report = Report()
        pre = {}
        if self.session is not None:
            pre["Session"] = self.session.idx
        if self.idx is not None:
            pre["Run"] = self.idx
        if self.model:
            pre["Model"] = self.model.name
        if len(self.frontends) > 0:
            pre["Frontend"] = self.get_frontend_name()
        if self.framework:
            pre["Framework"] = self.framework.name
        if self.backend:
            pre["Backend"] = self.backend.name
        if len(self.platforms) > 0:
            pre["Platform"] = self.get_platform_name()
        if self.target:
            pre["Target"] = self.target.name
        post = {}
        post["Features"] = self.get_all_feature_names()
        # post["Config"] = self.get_all_configs(omit_paths=True, omit_defaults=True, omit_globals=True)
        post["Config"] = self.get_all_configs(omit_paths=True, omit_defaults=False, omit_globals=True)
        post["Postprocesses"] = self.get_all_postprocess_names()
        post["Comment"] = self.comment if len(self.comment) > 0 else "-"
        if self.failing:
            post["Failing"] = True
            reason_text = self.get_reason_text()
            if reason_text:
                post["Reason"] = reason_text

        self.export_stage(RunStage.RUN, optional=self.export_optional)

        subs = []
        # metrics = Metrics()
        metrics_by_sub = {}

        def metrics_helper(stage, subs):
            # if self.failing:
            #     return subs
            if stage in self.artifacts_per_stage:
                prev_metrics_by_sub = metrics_by_sub.copy()
                names = self.artifacts_per_stage[stage].keys()
                subs = names
                for name in self.artifacts_per_stage[stage]:
                    # metrics_by_sub[name] = Metrics()
                    if len(self.artifacts_per_stage[stage][name]) > 0:
                        filename = f"{stage.name.lower()}_metrics.csv"
                        metrics_artifact = lookup_artifacts(self.artifacts_per_stage[stage][name], name=filename)
                        if len(metrics_artifact) == 0:
                            continue
                        assert len(metrics_artifact) == 1
                        metrics_artifact = metrics_artifact[0]
                        metrics = Metrics.from_csv(metrics_artifact.content)
                        metrics_data = metrics.get_data(include_optional=self.export_optional)
                        # Combine with existing metrics
                        parents = self.sub_parents[(stage, name)]
                        parent_stage, parent_name = parents
                        if parent_name in prev_metrics_by_sub:
                            parent_metrics_data = prev_metrics_by_sub[parent_name].get_data(
                                include_optional=self.export_optional
                            )
                        else:
                            parent_metrics_data = {}
                        for key, value in parent_metrics_data.items():
                            if key not in metrics_data:
                                metrics.add(key, value)
                        metrics_by_sub[name] = metrics
            return subs

        for stage in range(RunStage.LOAD, RunStage.POSTPROCESS):
            subs_ = metrics_helper(RunStage(stage), subs)
            if len(subs_) < len(subs):
                # assert self.failing
                pass
            else:
                subs = subs_

        pres = []
        mains = []
        posts = []
        for sub in subs:
            if sub not in ["", "default"]:
                pres.append({**pre, "Sub": sub})
            else:
                pres.append(pre)
            if sub in metrics_by_sub:
                main = metrics_by_sub[sub].get_data(include_optional=self.export_optional)
            else:
                main = {}
            mains.append(main if len(main) > 0 else {"Incomplete": True})
            posts.append(post)  # TODO: omit for subs?
        report.set(pre=pres, main=mains, post=posts)
        return report

    def export(self, path=None, optional=False):
        """Write a run configuration to a disk."""
        logger.debug("%s Exporting run to disk", self.prefix)
        if path is not None:
            raise NotImplementedError
        assert not self.locked
        for stage in range(self.next_stage):
            if not self.has_stage(stage):
                continue
            self.export_stage(stage, optional=optional)

        self.write_run_file()


# TODO: implement close()? and use closing contextlib?



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os

from mlonmcu.target.target import Target
from mlonmcu.target.metrics import Metrics

from mlonmcu.logging import get_logger

logger = get_logger()


def name2device(name):
    return name.replace("tvm_", "")


def create_tvm_platform_target(name, platform, base=Target):
    class TvmPlatformTarget(base):
        DEFAULTS = {
            **base.DEFAULTS,
            "timeout_sec": 0,  # disabled
        }

        def __init__(self, features=None, config=None):
            super().__init__(name=name, features=features, config=config)
            self.platform = platform
            self.device = name2device(name)

        @property
        def timeout_sec(self):
            return int(self.config["timeout_sec"])

        @property
        def repeat(self):
            return None  # This is handled at the platform level

        def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
            """Use target to execute a executable with given arguments"""
            if len(args) > 0:
                raise RuntimeError("Program arguments are not supported for real hardware devices")

            assert self.platform is not None, "TVM targets need a platform to execute programs"

            if self.timeout_sec > 0:
                raise NotImplementedError

            ret = self.platform.run(program, self)
            return ret

        def parse_stdout(self, out):
            mean_ms = None
            median_ms = None
            max_ms = None
            min_ms = None
            std_ms = None
            found = False
            for line in out.split("\n"):
                if found:
                    match = re.compile(r"\s+(\d*\.\d+)\s+(\d*\.\d+)\s+(\d*\.\d+)\s+(\d*\.\d+)\s+(\d*\.\d+)").findall(
                        line
                    )
                    assert len(match) == 1
                    groups = match[0]
                    mean_ms, median_ms, max_ms, min_ms, std_ms = (
                        float(groups[0]),
                        float(groups[1]),
                        float(groups[2]),
                        float(groups[3]),
                        float(groups[4]),
                    )
                    break
                if re.compile(r"\s+mean \(ms\)\s+median \(ms\)\s+max \(ms\)\s+min \(ms\)\s+std \(ms\)").match(line):
                    found = True
            return mean_ms, median_ms, max_ms, min_ms, std_ms

        def get_metrics(self, elf, directory, handle_exit=None):
            if self.print_outputs:
                out = self.exec(elf, cwd=directory, live=True, handle_exit=handle_exit)
            else:
                out = self.exec(
                    elf,
                    cwd=directory,
                    live=False,
                    print_func=lambda *args, **kwargs: None,
                    handle_exit=handle_exit,
                )
            mean_ms, _, max_ms, min_ms, _ = self.parse_stdout(out)

            metrics = Metrics()
            mean_s = mean_ms / 1e3 if mean_ms is not None else mean_ms
            min_s = min_ms / 1e3 if min_ms is not None else min_ms
            max_s = max_ms / 1e3 if max_ms is not None else max_ms
            if (
                self.platform.number == 1
                and self.platform.repeat == 1
                and (not self.platform.total_time or self.platform.aggregate != "none")
            ):
                metrics.add("Runtime [s]", mean_s)
            else:
                if self.platform.total_time:
                    metrics.add("Total Runtime [s]", mean_s * self.platform.number)
                if self.platform.aggregate == "all":
                    metrics.add("Average Runtime [s]", mean_s)
                    metrics.add("Min Runtime [s]", min_s)
                    metrics.add("Max Runtime [s]", max_s)
                elif self.platform.aggregate in ["avg", "mean"]:
                    metrics.add("Average Runtime [s]", mean_s)
                elif self.platform.aggregate == "min":
                    metrics.add("Min Runtime [s]", min_s)
                elif self.platform.aggregate == "max":
                    metrics.add("Max Runtime [s]", max_s)

            if self.platform.profile:
                headers = None
                lines = out.split("\n")
                extracted = []

                def extract_cols(line):
                    x = re.compile(r"(([^\s\[\]]+)(\s\S+)*)(\[.*\])?").findall(line)
                    return [y[0] for y in x]

                for line in lines:
                    if "---" in line:
                        break
                    if headers is None:
                        if "Name" in line:
                            headers = extract_cols(line)
                    else:
                        cols = extract_cols(line)
                        data = {headers[i]: val for i, val in enumerate(cols)}
                        extracted.append(data)
                assert len(extracted) > 0
                metrics = {"default": metrics}
                for item in extracted:
                    metrics_ = Metrics()
                    metrics_.add("Runtime [s]", float(item["Duration (us)"]) / 1e6)
                    metrics[item["Name"]] = metrics_

            return metrics, out, []

        def get_arch(self):
            return "unkwown"

        def update_environment(self, env):
            # TODO: implement in base class?
            pass

    return TvmPlatformTarget



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Collection of (example) postprocesses integrated in MLonMCU."""

import re
import ast
import tempfile
from pathlib import Path
from io import StringIO

import pandas as pd

from mlonmcu.artifact import Artifact, ArtifactFormat, lookup_artifacts
from mlonmcu.config import str2dict, str2bool, str2list
from mlonmcu.logging import get_logger

from .postprocess import SessionPostprocess, RunPostprocess

logger = get_logger()


def match_rows(df, cols):
    """Helper function to group similar rows in a dataframe."""
    groups = df.astype(str).groupby(cols).apply(lambda x: tuple(x.index)).tolist()
    return groups


def _check_cfg(value):
    res = re.compile(r"^((?:[a-zA-Z\d\-_ \.\[\]]+)(?:,[a-zA-Z\d\-_ \.\[\]]+)*)$").match(value)
    if res is None:
        return False
    return True


def _parse_cfg(value):
    if _check_cfg(value):
        return value.split(",")
    else:
        return ast.literal_eval(value)


class FilterColumnsPostprocess(SessionPostprocess):
    """Postprocess which can be used to drop unwanted columns from a report."""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
        "keep": None,
        "drop": None,
        "drop_nan": False,
        "drop_empty": False,
        "drop_const": False,
    }

    def __init__(self, features=None, config=None):
        super().__init__("filter_cols", features=features, config=config)

    @property
    def keep(self):
        """Get keep property."""
        cfg = self.config["keep"]
        if isinstance(cfg, str):
            return _parse_cfg(cfg)
        return cfg

    @property
    def drop(self):
        """Get drop property."""
        cfg = self.config["drop"]
        if isinstance(cfg, str):
            return _parse_cfg(cfg)
        return cfg

    @property
    def drop_nan(self):
        """Get drop_nan property."""
        value = self.config["drop_nan"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def drop_empty(self):
        """Get drop_empty property."""
        value = self.config["drop_empty"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def drop_const(self):
        """Get drop_const property."""
        value = self.config["drop_const"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def post_session(self, report):
        """Called at the end of a session."""

        def _filter_df(df, keep, drop, drop_nan=False, drop_empty=False, drop_const=False):
            if drop_empty:
                raise NotImplementedError
            if drop_nan:
                df.dropna(axis=1, how="all", inplace=True)
            if drop_const:
                df = df.loc[:, (df != df.iloc[0]).any()]
            if not (keep is None or drop is None):
                raise RuntimeError("'drop' and 'keep' can not be defined at the same time")
            if keep is not None:
                drop_cols = [name for name in df.columns if name not in keep]
            elif drop is not None:
                drop_cols = [name for name in df.columns if name in drop]
            else:
                drop_cols = []
            return df.drop(columns=drop_cols)

        report.pre_df = _filter_df(
            report.pre_df,
            self.keep,
            self.drop,
            drop_nan=self.drop_nan,
            drop_empty=self.drop_empty,
            drop_const=self.drop_const,
        )
        report.main_df = _filter_df(
            report.main_df, self.keep, self.drop, drop_nan=self.drop_nan, drop_const=self.drop_const
        )
        report.post_df = _filter_df(
            report.post_df, self.keep, self.drop, drop_nan=self.drop_nan, drop_const=self.drop_const
        )


class RenameColumnsPostprocess(SessionPostprocess):
    """Postprocess which can rename columns based on a provided mapping."""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
        "mapping": {},
        "merge": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("rename_cols", features=features, config=config)

    @property
    def mapping(self):
        value = self.config["mapping"]
        if not isinstance(value, dict):
            return str2dict(value)
        return value

    @property
    def merge(self):
        value = self.config["merge"]
        return str2bool(value)

    def post_session(self, report):
        """Called at the end of a session."""
        values = self.mapping.values()
        if len(values) != len(set(values)) and not self.merge:
            logger.warning("rename_cols: non unique mapping found. use merge=True to avoid overwriting values.")

        def merge(df):
            if len(set(df.columns)) == len(df.columns):
                return df
            a = df.loc[:, ~df.columns.duplicated(keep="first")]
            b = df.loc[:, df.columns.duplicated(keep="first")]
            return a.combine_first(merge(b))

        report.pre_df = report.pre_df.rename(columns=self.mapping)
        report.main_df = report.main_df.rename(columns=self.mapping)
        report.post_df = report.post_df.rename(columns=self.mapping)

        if self.merge:
            report.pre_df = merge(report.pre_df)
            report.main_df = merge(report.main_df)
            report.post_df = merge(report.post_df)


class Features2ColumnsPostprocess(SessionPostprocess):  # RunPostprocess?
    """Postprocess which can be used to transform (explode) the 'Features' Column
    in a dataframe for easier filtering."""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
        "limit": [],
        "drop": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("features2cols", features=features, config=config)

    @property
    def limit(self):
        """Get limit property."""
        value = self.config["limit"]
        if not isinstance(value, list):
            return str2list(value)
        return value

    @property
    def drop(self):
        """Get drop property."""
        value = self.config["drop"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def post_session(self, report):
        df = report.post_df
        if "Features" not in df.columns:
            return
        to_concat = [
            df["Features"].apply(lambda x: pd.Series({"feature_" + feature_name: feature_name in x}))
            for feature_name in list(set(df["Features"].sum()))
            if feature_name in self.limit or len(self.limit) == 0
        ]
        if len(to_concat) == 0:
            return
        feature_df = pd.concat(
            to_concat,
            axis=1,
        )
        if self.drop:
            tmp_df = df.drop(columns=["Features"])
        else:
            tmp_df = df
        new_df = pd.concat([tmp_df, feature_df], axis=1)
        report.post_df = new_df


class Config2ColumnsPostprocess(SessionPostprocess):  # RunPostprocess?
    """Postprocess which can be used to transform (explode) the 'Config' Column in a dataframe for easier filtering."""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
        "limit": [],
        "drop": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("config2cols", features=features, config=config)

    @property
    def limit(self):
        """Get limit property."""
        value = self.config["limit"]
        if not isinstance(value, list):
            return str2list(value)
        return value

    @property
    def drop(self):
        """Get drop property."""
        value = self.config["drop"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def post_session(self, report):
        """Called at the end of a session."""
        df = report.post_df
        if "Config" not in df.columns:
            return
        config_df = (
            df["Config"]
            .apply(lambda x: {key: value for key, value in x.items() if key in self.limit or len(self.limit) == 0})
            .apply(pd.Series)
            .add_prefix("config_")
        )
        if self.drop:
            tmp_df = df.drop(columns=["Config"])
        else:
            tmp_df = df
        new_df = pd.concat([tmp_df, config_df], axis=1)
        report.post_df = new_df


class MyPostprocess(SessionPostprocess):
    """TODO"""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
    }

    def __init__(self, features=None, config=None):
        super().__init__("mypost", features=features, config=config)
        self.config2cols = Config2ColumnsPostprocess(
            config={
                "config2cols.limit": [
                    "tvmllvm.desired_layout",
                    "tvmaot.desired_layout",
                    "tvmaotplus.desired_layout",
                    "tvmrt.desired_layout",
                    "xcorev.mem",
                    "xcorev.mac",
                    "xcorev.bi",
                    "xcorev.alu",
                    "xcorev.bitmanip",
                    "xcorev.simd",
                    "xcorev.hwlp",
                    "cv32e40p.fpu",
                    "etiss.fpu",
                    "corev_ovpsim.fpu",
                    "tvmaot.disabled_passes",
                    "tvmaotplus.disabled_passes",
                    "tvmrt.disabled_passes",
                    "tvmllvm.disabled_passes",
                    "auto_vectorize.loop",
                    "auto_vectorize.slp",
                    "auto_vectorize.force_vector_width",
                    "auto_vectorize.force_vector_interleave",
                    "auto_vectorize.custom_unroll",
                    "tvmllvm.target_keys",
                    "tvmrt.target_keys",
                    "tvmaot.target_keys",
                    "tvmaotplus.target_keys",
                    "autotuned.mode",
                ],
                "config2cols.drop": True,
            }
        )
        self.rename_cols = RenameColumnsPostprocess(
            config={
                "rename_cols.mapping": {
                    "config_tvmllvm.desired_layout": "Layout",
                    "config_tvmaot.desired_layout": "Layout",
                    "config_tvmaotplus.desired_layout": "Layout",
                    "config_tvmrt.desired_layout": "Layout",
                    "config_xcorev.mem": "XCVMem",
                    "config_xcorev.mac": "XCVMac",
                    "config_xcorev.bi": "XCVBi",
                    "config_xcorev.alu": "XCVAlu",
                    "config_xcorev.bitmanip": "XCVBitmanip",
                    "config_xcorev.simd": "XCVSimd",
                    "config_xcorev.hwlp": "XCVHwlp",
                    "feature_autotuned": "Autotuned",
                    "feature_debug": "Debug",
                    "config_cv32e40p.fpu": "FPU",
                    "config_etiss.fpu": "FPU",
                    "config_corev_ovpsim.fpu": "FPU",
                    "config_tvmaot.disabled_passes": "Disabled",
                    "config_tvmaotplus.disabled_passes": "Disabled",
                    "config_tvmrt.disabled_passes": "Disabled",
                    "config_tvmllvm.disabled_passes": "Disabled",
                    "config_auto_vectorize.loop": "Loop",
                    "config_auto_vectorize.slp": "Slp",
                    "config_auto_vectorize.force_vector_width": "FVW",
                    "config_auto_vectorize.force_vector_interleave": "FVI",
                    "config_auto_vectorize.custom_unroll": "Unroll",
                    "config_tvmllvm.target_keys": "Keys",
                    "config_tvmrt.target_keys": "Keys",
                    "config_tvmaot.target_keys": "Keys",
                    "config_tvmaotplus.target_keys": "Keys",
                    "config_autotuned.mode": "Tuner",
                }
            }
        )
        self.features2cols = Features2ColumnsPostprocess(
            config={
                "features2cols.limit": ["autotuned", "debug", "auto_vectorize", "target_optimized"],
                "features2cols.drop": True,
            }
        )
        self.filter_cols = FilterColumnsPostprocess(
            config={
                "filter_cols.drop": [
                    "Postprocesses",
                    "Framework",
                    "Platform",
                    "Session",
                    "ROM read-only",
                    "ROM code",
                    "ROM misc",
                    "RAM data",
                    "RAM zero-init data",
                    "Run Stage Time [s]",
                    "Compile Stage Time [s]",
                    "Workspace Size [B]",
                    "Build Stage Time [s]",
                    "Load Stage Time [s]",
                    "feature_auto_vectorize",
                    "feature_target_optimized",
                    "Setup Cycles",
                    "Setup Instructions",
                    "Setup CPI",
                ]
            }
        )

    def post_session(self, report):
        """TODO"""
        self.config2cols.post_session(report)
        self.features2cols.post_session(report)
        self.rename_cols.post_session(report)
        self.filter_cols.post_session(report)


class PassConfig2ColumnsPostprocess(SessionPostprocess):
    """Postprocess which can be used to transform (explode) the TVM pass_config into separate columns.
    requires prior Config2Columns pass."""

    def __init__(self, features=None, config=None):
        super().__init__("passcfg2cols", features=features, config=config)

    def post_session(self, report):
        """Called at the end of a session."""
        df = report.post_df
        name = "config_tvmaot.extra_pass_config"
        if name not in df.columns:
            return
        config_df = df[name].apply(pd.Series).add_prefix("passcfg_")
        tmp_df = df.drop(columns=[name])
        new_df = pd.concat([tmp_df, config_df], axis=1)
        report.post_df = new_df


class Bytes2kBPostprocess(SessionPostprocess):  # RunPostprocess?
    """Postprocess which can be used to scale the memory related columns from Bytes to KiloBytes."""

    def __init__(self, features=None, config=None):
        super().__init__("bytes2kb", features=features, config=config)

    def post_session(self, report):
        """Called at the end of a session."""
        df = report.main_df
        match_strs = ["ROM", "RAM"]
        cols = list(
            filter(lambda x: any(s in x for s in match_strs), df.columns)
        )  # Only scale columns related to memory
        cols = [col for col in cols if "kB" not in col]  # Do not scale columns with are already in kB

        for col in cols:
            df[col] = df[col] / 1000.0
            df.rename(columns={col: col + " [kB]"}, inplace=True)

        report.main_df = df


class VisualizePostprocess(SessionPostprocess):
    """A very simple example on how to generate a plot of the results using a postprocess."""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
        "format": "png",
    }

    def __init__(self, features=None, config=None):
        super().__init__("visualize", features=features, config=config)

    @property
    def format(self):
        """Get format property."""
        return self.config["format"]

    def post_session(self, report):
        """Called at the end of a session."""
        df = pd.concat([report.pre_df, report.main_df], axis=1)

        if self.format != "png":
            raise NotImplementedError("Currently only supports PNG")

        COLS = ["Cycles", "Total ROM", "Total RAM"]
        for col in COLS:
            if col not in report.main_df.columns:
                return []

        # Local import to deal with optional dependencies
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(ncols=len(COLS))
        plt.rcParams["figure.figsize"] = (15, 3)  # (w, h)
        for i, col in enumerate(COLS):
            new_df = df[[col]].astype(float)
            bar_names_df = (
                df["Session"].astype(str) + "_" + df["Run"].astype(str)
            )  # ideally we would use model/backend/target names here...
            new_df.index = bar_names_df
            new_df.plot(kind="bar", ax=axes[i])

        data = None
        with tempfile.TemporaryDirectory() as tmpdirname:
            fig_path = Path(tmpdirname) / "plot.png"
            fig.savefig(fig_path)
            with open(fig_path, "rb") as handle:
                data = handle.read()

        artifacts = [Artifact("plot.png", raw=data, fmt=ArtifactFormat.RAW)]
        return artifacts


class Artifact2ColumnPostprocess(RunPostprocess):
    """Postprocess for converting artifacts to columns in the report."""

    DEFAULTS = {
        **RunPostprocess.DEFAULTS,
        "file2colname": {},
    }

    def __init__(self, features=None, config=None):
        super().__init__("artifacts2cols", features=features, config=config)

    @property
    def file2colname(self):
        """Get file2colname property."""
        value = self.config["file2colname"]
        if not isinstance(value, dict):
            return str2dict(value)
        return value

    def post_run(self, report, artifacts):
        """Called at the end of a run."""
        for filename, colname in self.file2colname.items():
            filename = Path(filename)
            filecol = None
            if ":" in filename.name:
                fname, filecol = filename.name.rsplit(":", 1)
                filename = filename.parent / fname
            matches = lookup_artifacts(artifacts, name=filename, first_only=True)
            if not matches:
                report.main_df[colname] = ""
                continue
            if matches[0].fmt != ArtifactFormat.TEXT:
                raise RuntimeError("Can only put text into report columns")
            content = matches[0].content
            if filecol:
                assert filename.suffix == ".csv"
                filedf = pd.read_csv(StringIO(content))
                if filecol == "*":
                    cols = list(filedf.columns)
                else:
                    assert filecol in filedf.columns
                    cols = [filecol]
                content = filedf[cols].to_dict(orient="list")
                if len(content) == 1:
                    content = content[list(content.keys())[0]]
                if len(content) == 1:
                    content = content[0]
                content = str(content)
            report.main_df[colname] = content
        return []


class AnalyseInstructionsPostprocess(RunPostprocess):
    """Counting specific types of instructions."""

    DEFAULTS = {
        **RunPostprocess.DEFAULTS,
        "groups": True,
        "sequences": True,
        "seq_depth": 3,
        "top": 10,
        "to_df": False,
        "to_file": True,
        "corev": False,
    }

    def __init__(self, features=None, config=None):
        super().__init__("analyse_instructions", features=features, config=config)

    @property
    def groups(self):
        """Get groups property."""
        value = self.config["groups"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def sequences(self):
        """get sequences property."""
        value = self.config["sequences"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def seq_depth(self):
        """get seq_depth property."""
        return int(self.config["seq_depth"])

    @property
    def top(self):
        """get top property."""
        return int(self.config["top"])

    @property
    def to_df(self):
        """Get to_df property."""
        value = self.config["to_df"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def to_file(self):
        """Get to_file property."""
        value = self.config["to_file"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def corev(self):
        """Get corev property."""
        value = self.config["corev"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def post_run(self, report, artifacts):
        """Called at the end of a run."""
        ret_artifacts = []
        log_artifact = lookup_artifacts(artifacts, flags=("log_instrs",), fmt=ArtifactFormat.TEXT, first_only=True)
        assert len(log_artifact) == 1, "To use analyse_instructions process, please enable feature log_instrs."
        log_artifact = log_artifact[0]
        is_spike = "spike" in log_artifact.flags
        is_etiss = "etiss_pulpino" in log_artifact.flags or "etiss" in log_artifact.flags
        is_ovpsim = "ovpsim" in log_artifact.flags or "corev_ovpsim" in log_artifact.flags
        is_riscv = is_spike or is_etiss or is_ovpsim
        if is_spike:
            content = log_artifact.content
            if self.groups:
                encodings = re.compile(r"\((0x[0-9abcdef]+)\)").findall(content)
            if self.sequences:
                names = re.compile(r"core\s+\d+:\s0x[0-9abcdef]+\s\(0x[0-9abcdef]+\)\s([\w.]+).*").findall(content)
        elif is_etiss:
            content = log_artifact.content
            if self.groups:
                encodings = re.compile(r"0x[0-9abcdef]+:\s\w+\s#\s([0-9a-fx]+)\s.*").findall(content)
                encodings = [f"0b{enc}" for enc in encodings]
                # encodings = [f"{enc}" for enc in encodings]
            if self.sequences:
                names = re.compile(r"0x[0-9abcdef]+:\s(\w+)\s#\s[0-9a-fx]+\s.*").findall(content)
        elif is_ovpsim:
            content = log_artifact.content
            if self.groups:
                encodings = re.compile(r"riscvOVPsim\/cpu',\s0x[0-9abcdef]+\(.*\):\s([0-9abcdef]+)\s+\w+\s+.*").findall(
                    content
                )
                encodings = [f"0x{enc}" for enc in encodings]
            if self.sequences:
                names = re.compile(r"riscvOVPsim\/cpu',\s0x[0-9abcdef]+\(.*\):\s[0-9abcdef]+\s+(\w+)\s+.*").findall(
                    content
                )
        else:
            raise RuntimeError("Uable to determine the used target.")

        def _helper(x, top=100):
            counts = pd.Series(x).value_counts()
            probs = counts / len(x)
            return dict(counts.head(top)), dict(probs.head(top))

        def _gen_csv(label, counts, probs):
            lines = [f"{label},Count,Probablity"]
            for x in counts:
                line = f"{x},{counts[x]},{probs[x]:.3f}"
                lines.append(line)
            return "\n".join(lines)

        if self.groups:
            assert is_riscv, "Currently only riscv instrcutions can be analysed by groups"

            def _extract_major_opcode(enc):
                mapping = {
                    0b0010011: "OP-IMM",
                    0b0110111: "LUI",
                    0b0010111: "AUIPC",
                    0b0110011: "OP",
                    0b1101111: "JAL",
                    0b1100111: "JALR",
                    0b1100011: "BRANCH",
                    0b0000011: "LOAD",
                    0b0100011: "STORE",
                    0b0001111: "MISC-MEM",
                    0b1110011: "SYSTEM",
                    0b1000011: "MADD",
                    0b1000111: "MSUB",
                    0b1001011: "MNSUB",
                    0b1001111: "MNADD",
                    0b0000111: "LOAD-FP",
                    0b0100111: "STORE-FP",
                    0b0001011: "custom-0",
                    0b0101011: "custom-1",
                    0b1011011: "custom-2/rv128",
                    0b1111011: "custom-3/rv128",
                    0b1101011: "reserved",
                    0b0101111: "AMO",
                    0b1010011: "OP-FP",
                    0b1010111: "OP-V",
                    0b1110111: "OP-P",
                    0b0011011: "OP-IMM-32",
                    0b0111011: "OP-32",
                }
                enc = int(enc, 0)  # Convert from hexadecimal
                opcode = enc & 0b1111111
                lsbs = opcode & 0b11
                if lsbs == 0b11:
                    major = mapping.get(opcode, "UNKNOWN")
                else:
                    # 16-bit instruction
                    msbs = (enc & 0b1110000000000000) >> 13
                    rvc_mapping = {
                        0b00000: "OP-IMM",
                        0b00001: "OP-IMM",
                        0b00010: "OP-IMM",
                        0b00100: "LOAD",
                        0b00101: "JAL",
                        0b00110: "LOAD-FP",
                        0b01000: "LOAD",
                        0b01001: "OP-IMM",
                        0b01010: "LOAD",
                        0b01100: "LOAD-FP",
                        0b01101: "OP-IMM",
                        0b01110: "LOAD-FP",
                        0b10000: "reserved",
                        0b10001: "MISC-ALU",
                        0b10010: "JALR",
                        0b10100: "STORE-FP",
                        0b10101: "JAL",
                        0b10110: "STORE-FP",
                        0b11000: "STORE",
                        0b11001: "BRANCH",
                        0b11010: "STORE",
                        0b11100: "STORE-FP",
                        0b11101: "BRANCH",
                        0b11110: "STORE-FP",
                    }
                    combined = msbs << 2 | lsbs
                    assert combined in rvc_mapping.keys()
                    return f"{rvc_mapping[combined]} (Compressed)"
                return major

            majors = list(map(_extract_major_opcode, encodings))
            major_counts, major_probs = _helper(majors, top=self.top)
            majors_csv = _gen_csv("Major", major_counts, major_probs)
            artifact = Artifact("analyse_instructions_majors.csv", content=majors_csv, fmt=ArtifactFormat.TEXT)
            if self.to_file:
                ret_artifacts.append(artifact)
            if self.to_df:
                post_df = report.post_df.copy()
                post_df["AnalyseInstructionsMajorsCounts"] = str(major_counts)
                post_df["AnalyseInstructionsMajorsProbs"] = str(major_probs)
                report.post_df = post_df
        if self.sequences:
            max_len = self.seq_depth

            def _get_sublists(lst, length):
                ret = []
                for i in range(len(lst) - length + 1):
                    lst_ = lst[i : i + length]
                    ret.append(";".join(lst_))
                return ret

            for length in range(1, max_len + 1):
                names_ = _get_sublists(names, length)
                counts, probs = _helper(names_, top=self.top)
                sequence_csv = _gen_csv("Sequence", counts, probs)
                artifact = Artifact(
                    f"analyse_instructions_seq{length}.csv", content=sequence_csv, fmt=ArtifactFormat.TEXT
                )
                if self.to_file:
                    ret_artifacts.append(artifact)
                if self.to_df:
                    post_df = report.post_df.copy()
                    post_df[f"AnalyseInstructionsSeq{length}Counts"] = str(counts)
                    post_df[f"AnalyseInstructionsSeq{length}Probs"] = str(probs)
                    report.post_df = post_df
        if self.corev:
            XCVMAC_INSNS = {
                "cv.mac",
                "cv.msu",
                "cv.mulun",
                "cv.mulhhun",
                "cv.mulsn",
                "cv.mulhhsn",
                "cv.mulurn",
                "cv.mulhhurn",
                "cv.mulsrn",
                "cv.mulhhsrn",
                "cv.macun",
                "cv.machhun",
                "cv.macsn",
                "cv.machhsn",
                "cv.macurn",
                "cv.machhurn",
                "cv.macsrn",
                "cv.machhsrn",
            }
            XCVMEM_INSNS = {
                "cv.lb_ri_inc",
                "cv.lbu_ri_inc",
                "cv.lh_ri_inc",
                "cv.lhu_ri_inc",
                "cv.lw_ri_inc",
                "cv.lb_ri_inc",
                "cv.lbu_ri_inc",
                "cv.lh_ri_inc",
                "cv.lhu_ri_inc",
                "cv.lw_ri_inc",
                "cv.lb_rr_inc",
                "cv.lbu_rr_inc",
                "cv.lh_rr_inc",
                "cv.lhu_rr_inc",
                "cv.lw_rr_inc",
                "cv.lb_rr_inc",
                "cv.lbu_rr_inc",
                "cv.lh_rr_inc",
                "cv.lhu_rr_inc",
                "cv.lw_rr_inc",
                "cv.lb_rr",
                "cv.lbu_rr",
                "cv.lh_rr",
                "cv.lhu_rr",
                "cv.lw_rr",
                "cv.sb_ri_inc",
                "cv.sh_ri_inc",
                "cv.sw_ri_inc",
                "cv.sb_ri_inc",
                "cv.sh_ri_inc",
                "cv.sw_ri_inc",
                "cv.sb_rr_inc",
                "cv.sh_rr_inc",
                "cv.sw_rr_inc",
                "cv.sb_rr_inc",
                "cv.sh_rr_inc",
                "cv.sw_rr_inc",
                "cv.sb_rr",
                "cv.sh_rr",
                "cv.sw_rr",
            }
            XCVBI_INSNS = {
                "cv.bneimm",
                "cv.beqimm",
            }
            XCVALU_INSNS = {
                "cv.slet",
                "cv.min",
                "cv.addnr",
                "cv.addunr",
                "cv.addn",
                "cv.maxu",
                "cv.subun",
                "cv.extbz",
                "cv.addun",
                "cv.clip",
                "cv.clipu",
                "cv.subn",
                "cv.max",
                "cv.extbs",
                "cv.abs",
                "cv.addurn",
                "cv.exths",
                "cv.exthz",
                "cv.minu",
                "cv.sletu",
                "cv.suburn",
                "cv.addrn",
                "cv.clipur",
                "cv.subrn",
            }
            XCVBITMANIP_INSNS = {
                "cv.ror",
                "cv.clb",
            }
            XCVSIMD_INSNS = {
                "cv.add.h",
                "cv.add.sc.b",
                "cv.add.sc.h",
                "cv.add.sci.h",
                "cv.and.b",
                "cv.and.h",
                "cv.and.sc.h",
                "cv.and.sci.h",
                "cv.cmpeq.sc.h",
                "cv.cmpge.sci.h",
                "cv.cmpgtu.h",
                "cv.cmplt.sci.h",
                "cv.cmpltu.sci.b",
                "cv.cmpne.sc.h",
                "cv.cmpne.sci.b",
                "cv.extract.b",
                "cv.extract.h",
                "cv.extractu.b",
                "cv.extractu.h",
                "cv.insert.h",
                "cv.max.h",
                "cv.max.sci.h",
                "cv.maxu.h",
                "cv.or.b",
                "cv.or.h",
                "cv.pack",
                "cv.packhi.b",
                "cv.packlo.b",
                "cv.shuffle2.b",
                "cv.shuffle2.h",
                "cv.shufflei0.sci.b",
                "cv.sll.sci.h",
                "cv.sra.h",
                "cv.sra.sci.h",
                "cv.srl.h",
                "cv.srl.sci.h",
                "cv.sub.b",
                "cv.sub.h",
                "cv.xor.b",
                "cv.xor.sci.b",
                "cv.add.sci.b",
                "cv.cmpeq.b",
                "cv.cmpgtu.sc.h",
                "cv.cmpleu.sc.h",
                "cv.sdotup.h",
                "cv.sdotup.b",
                "cv.shuffle.sci.h",
                "cv.xor.sc.b",
                "cv.xor.sc.h",
                "cv.sdotsp.h",
                "cv.cmpeq.sci.b",
                "cv.and.sci.b",
                "cv.dotsp.h",
                "cv.dotsp.b",
                "cv.sdotsp.b",
                "cv.add.b",
                "cv.dotup.sci.b",
            }
            XCVHWLP_INSNS = {
                "cv.count",
                "cv.counti",
                "cv.start",
                "cv.starti",
                "cv.end",
                "cv.endi",
                "cv.setup",
                "cv.setupi",
            }

            def apply_mapping(x):
                x = x.replace("cv_", "cv.")
                x = x.replace("_sc", ".sc")
                x = x.replace("_b", ".b")
                x = x.replace("_h", ".h")
                if x in XCVMAC_INSNS:
                    return "XCVMac"
                elif x in XCVMEM_INSNS:
                    return "XCVMem"
                elif x in XCVALU_INSNS:
                    return "XCVAlu"
                elif x in XCVBITMANIP_INSNS:
                    return "XCVBitmanip"
                elif x in XCVBI_INSNS:
                    return "XCVBi"
                elif x in XCVSIMD_INSNS:
                    return "XCVSimd"
                elif x in XCVHWLP_INSNS:
                    return "XCVHwlp"
                elif "cv." in x:
                    return "XCV?"
                else:
                    return "Other"

            names_ = list(map(apply_mapping, names))
            cv_ext_counts, cv_ext_probs = _helper(names_, top=self.top)
            corev_csv = _gen_csv("Set", cv_ext_counts, cv_ext_probs)
            artifact = Artifact("analyse_instructions_corev.csv", content=corev_csv, fmt=ArtifactFormat.TEXT)
            if self.to_file:
                ret_artifacts.append(artifact)
            if self.to_df:
                post_df = report.post_df.copy()
                post_df["CoreVSetCounts"] = str(cv_ext_counts)
                post_df["CoreVSetProbs"] = str(cv_ext_probs)
                report.post_df = post_df
        assert self.to_file or self.to_df, "Either to_file or to_df have to be true"
        return ret_artifacts


class CompareRowsPostprocess(SessionPostprocess):
    """TODO"""

    DEFAULTS = {
        **SessionPostprocess.DEFAULTS,
        "to_compare": None,
        "group_by": None,
        "baseline": 0,
        "percent": False,
        "invert": False,
        "substract": False,
    }

    def __init__(self, features=None, config=None):
        super().__init__("compare_rows", features=features, config=config)

    @property
    def to_compare(self):
        """Get to_compare property."""
        value = self.config["to_compare"]
        return str2list(value, allow_none=True)

    @property
    def group_by(self):
        """Get group_by property."""
        value = self.config["group_by"]
        return str2list(value, allow_none=True)

    @property
    def baseline(self):
        """Get baseline property."""
        value = self.config["baseline"]
        return int(value)

    @property
    def percent(self):
        """Get percent property."""
        value = self.config["percent"]
        return str2bool(value)

    @property
    def invert(self):
        """Get invert property."""
        value = self.config["invert"]
        return str2bool(value)

    @property
    def substract(self):
        """Get substract property."""
        value = self.config["substract"]
        return str2bool(value)

    def post_session(self, report):
        """Called at the end of a session."""
        pre_df = report.pre_df
        main_df = report.main_df  # metrics
        post_df = report.post_df
        group_by = self.group_by
        if group_by is None:
            group_by = [x for x in pre_df.columns if x != "Run"]
        assert isinstance(group_by, list)
        assert all(col in list(pre_df.columns) + list(post_df.columns) for col in group_by), "Cols mssing in df"
        to_compare = self.to_compare
        if to_compare is None:
            to_compare = list(main_df.columns)
        assert isinstance(to_compare, list)
        assert all(col in main_df.columns for col in to_compare)
        full_df = pd.concat([pre_df, main_df, post_df], axis=1)
        grouped = full_df.groupby(group_by, axis=0, group_keys=False, dropna=False)
        new_df = pd.DataFrame()
        for col in to_compare:

            def f(df):
                assert self.baseline < len(df), "Index of group baseline out of bounds"
                ret = df / df.iloc[self.baseline]
                if self.substract:
                    ret = ret - 1
                if self.invert:
                    ret = 1 / ret
                if self.percent:
                    ret = ret * 100.0
                return ret

            filtered_col = grouped[col]
            first = filtered_col.apply(f).reset_index()
            first_col = first[col]
            new = first_col
            new_name = f"{col} (rel.)"
            new_df[new_name] = new
        main_df = pd.concat([main_df, new_df], axis=1)
        report.main_df = main_df


class AnalyseDumpPostprocess(RunPostprocess):
    """Counting static instructions."""

    DEFAULTS = {
        **RunPostprocess.DEFAULTS,
        "to_df": False,
        "to_file": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("analyse_dump", features=features, config=config)

    @property
    def to_df(self):
        """Get to_df property."""
        value = self.config["to_df"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def to_file(self):
        """Get to_file property."""
        value = self.config["to_file"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def post_run(self, report, artifacts):
        """Called at the end of a run."""
        platform = report.pre_df["Platform"]
        if (platform != "mlif").any():
            return []
        ret_artifacts = []
        dump_artifact = lookup_artifacts(
            artifacts, name="generic_mlonmcu.dump", fmt=ArtifactFormat.TEXT, first_only=True
        )
        assert len(dump_artifact) == 1, "To use analyse_dump postprocess, please set mlif.enable_asmdump=1"
        dump_artifact = dump_artifact[0]
        is_llvm = "llvm" in dump_artifact.flags
        assert is_llvm, "Non-llvm objdump currently unsupported"
        content = dump_artifact.content
        lines = content.split("\n")
        counts = {}
        total = 0
        for line in lines:
            splitted = line.split("\t")
            if len(splitted) != 3:
                continue
            insn = splitted[1]
            args = splitted[2]
            if "cv." in insn:
                if "(" in args and ")" in args:
                    m = re.compile(r"(.*)\((.*)\)").match(args)
                    if m:
                        g = m.groups()
                        assert len(g) == 2
                        offset, base = g
                        fmt = "ri"
                        try:
                            offset = int(offset)
                        except ValueError:
                            fmt = "rr"
                        insn += f"_{fmt}"
                        if "!" in base:
                            insn += "_inc"
            if insn in counts:
                counts[insn] += 1
            else:
                counts[insn] = 1
            total += 1
        counts_csv = "Instruction,Count,Probability\n"
        for insn, count in sorted(counts.items(), key=lambda item: item[1]):
            counts_csv += f"{insn},{count},{count/total:.4f}\n"
        artifact = Artifact("dump_counts.csv", content=counts_csv, fmt=ArtifactFormat.TEXT)
        if self.to_file:
            ret_artifacts.append(artifact)
        if self.to_df:
            post_df = report.post_df.copy()
            post_df["DumpCounts"] = str(counts)
            report.post_df = post_df
        assert self.to_file or self.to_df, "Either to_file or to_df have to be true"
        return ret_artifacts


class AnalyseCoreVCountsPostprocess(RunPostprocess):
    """Counting static instructions."""

    DEFAULTS = {
        **RunPostprocess.DEFAULTS,
        "to_df": False,
        "to_file": True,
    }

    def __init__(self, features=None, config=None):
        super().__init__("analyse_corev_counts", features=features, config=config)

    @property
    def to_df(self):
        """Get to_df property."""
        value = self.config["to_df"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def to_file(self):
        """Get to_file property."""
        value = self.config["to_file"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def post_run(self, report, artifacts):
        """Called at the end of a run."""
        ret_artifacts = []
        count_artifact = lookup_artifacts(artifacts, name="dump_counts.csv", fmt=ArtifactFormat.TEXT, first_only=True)
        assert len(count_artifact) == 1, "To use analyse_corev_counts postprocess, analyse_dump needs to run first."
        count_artifact = count_artifact[0]
        content = count_artifact.content

        lines = content.split("\n")

        XCVMAC_INSNS = {
            "cv.mac",
            "cv.msu",
            "cv.mulun",
            "cv.mulhhun",
            "cv.mulsn",
            "cv.mulhhsn",
            "cv.mulurn",
            "cv.mulhhurn",
            "cv.mulsrn",
            "cv.mulhhsrn",
            "cv.macun",
            "cv.machhun",
            "cv.macsn",
            "cv.machhsn",
            "cv.macurn",
            "cv.machhurn",
            "cv.macsrn",
            "cv.machhsrn",
        }
        XCVMEM_INSNS = {
            "cv.lb_ri_inc",
            "cv.lbu_ri_inc",
            "cv.lh_ri_inc",
            "cv.lhu_ri_inc",
            "cv.lw_ri_inc",
            "cv.lb_ri_inc",
            "cv.lbu_ri_inc",
            "cv.lh_ri_inc",
            "cv.lhu_ri_inc",
            "cv.lw_ri_inc",
            "cv.lb_rr_inc",
            "cv.lbu_rr_inc",
            "cv.lh_rr_inc",
            "cv.lhu_rr_inc",
            "cv.lw_rr_inc",
            "cv.lb_rr_inc",
            "cv.lbu_rr_inc",
            "cv.lh_rr_inc",
            "cv.lhu_rr_inc",
            "cv.lw_rr_inc",
            "cv.lb_rr",
            "cv.lbu_rr",
            "cv.lh_rr",
            "cv.lhu_rr",
            "cv.lw_rr",
            "cv.sb_ri_inc",
            "cv.sh_ri_inc",
            "cv.sw_ri_inc",
            "cv.sb_ri_inc",
            "cv.sh_ri_inc",
            "cv.sw_ri_inc",
            "cv.sb_rr_inc",
            "cv.sh_rr_inc",
            "cv.sw_rr_inc",
            "cv.sb_rr_inc",
            "cv.sh_rr_inc",
            "cv.sw_rr_inc",
            "cv.sb_rr",
            "cv.sh_rr",
            "cv.sw_rr",
        }
        XCVBI_INSNS = {
            "cv.bneimm",
            "cv.beqimm",
        }
        XCVALU_INSNS = {
            "cv.slet",
            "cv.min",
            "cv.addnr",
            "cv.addunr",
            "cv.addn",
            "cv.maxu",
            "cv.subun",
            "cv.extbz",
            "cv.addun",
            "cv.clip",
            "cv.clipu",
            "cv.subn",
            "cv.max",
            "cv.extbs",
            "cv.abs",
            "cv.addurn",
            "cv.exths",
            "cv.exthz",
            "cv.minu",
            "cv.sletu",
            "cv.suburn",
            "cv.addrn",
            "cv.clipur",
            "cv.subrn",
        }
        XCVBITMANIP_INSNS = {
            "cv.ror",
            "cv.clb",
        }
        XCVSIMD_INSNS = {
            "cv.add.h",
            "cv.add.sc.b",
            "cv.add.sc.h",
            "cv.add.sci.h",
            "cv.and.b",
            "cv.and.h",
            "cv.and.sc.h",
            "cv.and.sci.h",
            "cv.cmpeq.sc.h",
            "cv.cmpge.sci.h",
            "cv.cmpgtu.h",
            "cv.cmplt.sci.h",
            "cv.cmpltu.sci.b",
            "cv.cmpne.sc.h",
            "cv.cmpne.sci.b",
            "cv.extract.b",
            "cv.extract.h",
            "cv.extractu.b",
            "cv.extractu.h",
            "cv.insert.h",
            "cv.max.h",
            "cv.max.sci.h",
            "cv.maxu.h",
            "cv.or.b",
            "cv.or.h",
            "cv.pack",
            "cv.packhi.b",
            "cv.packlo.b",
            "cv.shuffle2.b",
            "cv.shuffle2.h",
            "cv.shufflei0.sci.b",
            "cv.sll.sci.h",
            "cv.sra.h",
            "cv.sra.sci.h",
            "cv.srl.h",
            "cv.srl.sci.h",
            "cv.sub.b",
            "cv.sub.h",
            "cv.xor.b",
            "cv.xor.sci.b",
            "cv.add.sci.b",
            "cv.cmpeq.b",
            "cv.cmpgtu.sc.h",
            "cv.cmpleu.sc.h",
            "cv.sdotup.h",
            "cv.sdotup.b",
            "cv.shuffle.sci.h",
            "cv.xor.sc.b",
            "cv.xor.sc.h",
            "cv.sdotsp.h",
            "cv.cmpeq.sci.b",
            "cv.and.sci.b",
            "cv.dotsp.h",
            "cv.dotsp.b",
            "cv.sdotsp.b",
            "cv.add.b",
            "cv.dotup.sci.b",
        }
        XCVHWLP_INSNS = {
            "cv.count",
            "cv.counti",
            "cv.start",
            "cv.starti",
            "cv.end",
            "cv.endi",
            "cv.setup",
            "cv.setupi",
        }

        unknowns = []
        cv_ext_totals = {
            "XCVMac": len(XCVMAC_INSNS),
            "XCVMem": len(XCVMEM_INSNS),
            "XCVBi": len(XCVBI_INSNS),
            "XCVAlu": len(XCVALU_INSNS),
            "XCVBitmanip": len(XCVBITMANIP_INSNS),
            "XCVSimd": len(XCVSIMD_INSNS),
            "XCVHwlp": len(XCVHWLP_INSNS),
            "Unknown": 0,
        }
        cv_ext_counts = {
            "XCVMac": 0,
            "XCVMem": 0,
            "XCVBi": 0,
            "XCVAlu": 0,
            "XCVBitmanip": 0,
            "XCVSimd": 0,
            "XCVHwlp": 0,
            "Unknown": 0,
        }
        cv_ext_unique_counts = {
            "XCVMac": 0,
            "XCVMem": 0,
            "XCVBi": 0,
            "XCVAlu": 0,
            "XCVBitmanip": 0,
            "XCVSimd": 0,
            "XCVHwlp": 0,
            "Unknown": 0,
        }
        total_counts = 0
        cv_counts_csv = "Instruction,Count,Probability\n"
        cv_counts = {}
        for line in lines[1:]:
            if "cv." not in line:
                continue
            cv_counts_csv += f"{line}\n"
            splitted = line.split(",")
            assert len(splitted) == 3
            insn = splitted[0]
            count = int(splitted[1])
            cv_counts[insn] = count
            total_counts += count
            if insn in XCVMAC_INSNS:
                cv_ext_counts["XCVMac"] += count
                cv_ext_unique_counts["XCVMac"] += 1
            elif insn in XCVMEM_INSNS:
                cv_ext_counts["XCVMem"] += count
                cv_ext_unique_counts["XCVMem"] += 1
            elif insn in XCVBI_INSNS:
                cv_ext_counts["XCVBi"] += count
                cv_ext_unique_counts["XCVBi"] += 1
            elif insn in XCVALU_INSNS:
                cv_ext_counts["XCVAlu"] += count
                cv_ext_unique_counts["XCVAlu"] += 1
            elif insn in XCVBITMANIP_INSNS:
                cv_ext_counts["XCVBitmanip"] += count
                cv_ext_unique_counts["XCVBitmanip"] += 1
            elif insn in XCVSIMD_INSNS:
                cv_ext_counts["XCVSimd"] += count
                cv_ext_unique_counts["XCVSimd"] += 1
            elif insn in XCVHWLP_INSNS:
                cv_ext_counts["XCVHwlp"] += count
                cv_ext_unique_counts["XCVHwlp"] += 1
            else:
                cv_ext_counts["Unknown"] += count
                cv_ext_unique_counts["Unknown"] += 1
                if insn not in unknowns:
                    unknowns.append(insn)
        cv_ext_totals["Unknown"] = len(unknowns)
        cv_ext_counts_csv = "Set,Count,Probability\n"
        for ext, count in sorted(cv_ext_counts.items(), key=lambda item: item[1]):
            if count == 0:
                continue
            cv_ext_counts_csv += f"{ext},{count},{count/total_counts}\n"
        cv_ext_unique_counts_csv = "Set,Used,Utilization\n"
        for ext, used in sorted(cv_ext_unique_counts.items(), key=lambda item: item[1]):
            if used == 0:
                continue
            rel = used / cv_ext_totals[ext]
            cv_ext_unique_counts_csv += f"{ext},{used},{rel:.4f}\n"
        used = sum(cv_ext_unique_counts.values())
        totals = sum(cv_ext_totals.values())
        rel = used / totals
        cv_ext_unique_counts_csv += f"XCVTotal,{used},{rel:.4f}\n"

        cv_counts_artifact = Artifact("cv_counts.csv", content=cv_counts_csv, fmt=ArtifactFormat.TEXT)
        cv_ext_counts_artifact = Artifact("cv_ext_counts.csv", content=cv_ext_counts_csv, fmt=ArtifactFormat.TEXT)
        cv_ext_unique_counts_artifact = Artifact(
            "cv_ext_unique_counts.csv", content=cv_ext_unique_counts_csv, fmt=ArtifactFormat.TEXT
        )
        if len(unknowns) > 0:
            logger.warning("Unknown instructions found: %s", unknowns)
            cv_ext_unknowns_artifact = Artifact(
                "cv_ext_unknowns.csv", content="\n".join(unknowns), fmt=ArtifactFormat.TEXT
            )
            if self.to_file:
                ret_artifacts.append(cv_ext_unknowns_artifact)
            # TODO: logging

        if self.to_file:
            ret_artifacts.append(cv_counts_artifact)
            ret_artifacts.append(cv_ext_counts_artifact)
            ret_artifacts.append(cv_ext_unique_counts_artifact)
        if self.to_df:
            post_df = report.post_df.copy()
            post_df["XCVCounts"] = str(cv_counts)
            post_df["XCVExtCounts"] = str(cv_ext_counts)
            post_df["XCVExtUniqueCounts"] = str(cv_ext_unique_counts)
            report.post_df = post_df
        assert self.to_file or self.to_df, "Either to_file or to_df have to be true"
        return ret_artifacts



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mlonmcu.cli.helper.filter import filter_arg


def test_cli_filter():
    assert filter_arg("x,y") == ["x", "y"]
    assert filter_arg("") == []
    assert filter_arg(None) == []



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os

from mlonmcu.target.target import Target
from mlonmcu.target.metrics import Metrics

from mlonmcu.logging import get_logger

from .microtvm_zephyr_target import ZephyrMicroTvmPlatformTarget
from .microtvm_arduino_target import ArduinoMicroTvmPlatformTarget
from .microtvm_espidf_target import EspidfMicroTvmPlatformTarget
from .microtvm_host_target import HostMicroTvmPlatformTarget
from .microtvm_etiss_target import EtissMicroTvmPlatformTarget
from .microtvm_spike_target import SpikeMicroTvmPlatformTarget
from .microtvm_corev_ovpsim_target import CoreVOVPSimMicroTvmPlatformTarget
from .microtvm_mlonmcu_target import MlonmcuMicroTvmPlatformTarget

logger = get_logger()


def name2template(name):
    return name.replace("microtvm_", "")


MICROTVM_PLATFORM_TARGET_REGISTRY = {}


def register_microtvm_platform_target(target_name, t, override=False):
    global MICROTVM_PLATFORM_TARGET_REGISTRY

    if target_name in MICROTVM_PLATFORM_TARGET_REGISTRY and not override:
        raise RuntimeError(f"MicroTVM platform target {target_name} is already registered")
    MICROTVM_PLATFORM_TARGET_REGISTRY[target_name] = t


def get_microtvm_platform_targets():
    return MICROTVM_PLATFORM_TARGET_REGISTRY


register_microtvm_platform_target("microtvm_zephyr", ZephyrMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_arduino", ArduinoMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_host", HostMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_etiss", EtissMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_espidf", EspidfMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_spike", SpikeMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_corev_ovpsim", CoreVOVPSimMicroTvmPlatformTarget)
register_microtvm_platform_target("microtvm_mlonmcu", MlonmcuMicroTvmPlatformTarget)


def create_microtvm_platform_target(name, platform, base=Target):
    class MicroTvmPlatformTarget(base):
        DEFAULTS = {
            **base.DEFAULTS,
            "timeout_sec": 0,  # disabled
        }

        def __init__(self, features=None, config=None):
            super().__init__(name=name, features=features, config=config)
            self.platform = platform
            self.template = name2template(name)

        @property
        def timeout_sec(self):
            return int(self.config["timeout_sec"])

        @property
        def repeat(self):
            return None  # This is handled at the platform level

        def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
            """Use target to execute a executable with given arguments"""
            if len(args) > 0:
                raise RuntimeError("Program arguments are not supported for real hardware devices")

            assert self.platform is not None, "TVM targets need a platform to execute programs"

            if self.timeout_sec > 0:
                raise NotImplementedError

            ret = self.platform.run(program, self)
            return ret

        def parse_stdout(self, out):
            mean_ms = None
            median_ms = None
            max_ms = None
            min_ms = None
            std_ms = None
            found = False
            for line in out.split("\n"):
                if found:
                    match = re.compile(r"\s+(\d*\.\d+)\s+(\d*\.\d+)\s+(\d*\.\d+)\s+(\d*\.\d+)\s+(\d*\.\d+)").findall(
                        line
                    )
                    assert len(match) == 1
                    groups = match[0]
                    mean_ms, median_ms, max_ms, min_ms, std_ms = (
                        float(groups[0]),
                        float(groups[1]),
                        float(groups[2]),
                        float(groups[3]),
                        float(groups[4]),
                    )
                    break
                if re.compile(r"\s+mean \(ms\)\s+median \(ms\)\s+max \(ms\)\s+min \(ms\)\s+std \(ms\)").match(line):
                    found = True
            return mean_ms, median_ms, max_ms, min_ms, std_ms

        def get_metrics(self, elf, directory, handle_exit=None):
            if self.print_outputs:
                out = self.exec(elf, cwd=directory, live=True, handle_exit=handle_exit)
            else:
                out = self.exec(
                    elf,
                    cwd=directory,
                    live=False,
                    print_func=lambda *args, **kwargs: None,
                    handle_exit=handle_exit,
                )
            mean_ms, _, _, _, _ = self.parse_stdout(out)

            metrics = Metrics()
            time_s = mean_ms / 1e3 if mean_ms is not None else mean_ms
            if time_s:
                metrics.add("Runtime [s]", time_s)

            if self.platform.profile:
                headers = None
                skip = False
                lines = out.split("\n")
                extracted = []

                def extract_cols(line):
                    x = re.compile(r"(([^\s\[\]]+)(\s\S+)*)(\[.*\])?").findall(line)
                    return [y[0] for y in x]

                for line in lines:
                    if skip:
                        skip = False
                        continue
                    if headers is None:
                        if "Name" in line:
                            headers = extract_cols(line)
                            skip = True
                            continue
                    else:
                        if len(line.strip()) == 0:
                            break
                        cols = extract_cols(line)
                        data = {headers[i]: val for i, val in enumerate(cols)}
                        extracted.append(data)
                        if "Total_time" in line:
                            break
                assert len(extracted) > 0
                metrics = {"default": metrics}
                for item in extracted:
                    if item["Node Name"] == "Total_time":
                        metrics["default"].add("Runtime [s]", float(item["Time(us)"]) / 1e6)
                    else:
                        metrics_ = Metrics()
                        metrics_.add("Runtime [s]", float(item["Time(us)"]) / 1e6)
                        metrics[item["Node Name"]] = metrics_

            return metrics, out, []

        def get_arch(self):
            return "unkwown"

    return MicroTvmPlatformTarget



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test target registry mechanism."""
import pytest

from mlonmcu.target import Target, register_target, get_targets


class MyTarget(Target):
    def __init__(self, name="foo", features=None, config=None):
        super().__init__(name, features=features, config=config)


def test_target_registry():
    before = len(get_targets())
    register_target("foo", MyTarget)
    register_target("bar", MyTarget)
    assert len(get_targets()) == before + 2
    with pytest.raises(RuntimeError):
        register_target("foo", MyTarget)
    assert len(get_targets()) == before + 2
    register_target("foo", MyTarget, override=True)
    assert len(get_targets()) == before + 2



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os

from mlonmcu.target.target import Target
from mlonmcu.target.metrics import Metrics

from mlonmcu.logging import get_logger

logger = get_logger()


def create_zephyr_platform_target(name, platform, base=Target):
    class ZephyrPlatformTarget(base):
        DEFAULTS = {
            **base.DEFAULTS,
            "timeout_sec": 0,  # disabled
            "port": None,
            "baud": None,
        }

        def __init__(self, features=None, config=None):
            super().__init__(name=name, features=features, config=config)
            self.platform = platform

        @property
        def timeout_sec(self):
            return int(self.config["timeout_sec"])

        @property
        def port(self):
            return self.config["port"]

        @property
        def baud(self):
            return self.config["baud"]

        def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
            """Use target to execute a executable with given arguments"""
            if len(args) > 0:
                raise RuntimeError("Program arguments are not supported for real hardware devices")

            assert self.platform is not None, "Zephyr targets need a platform to execute programs"

            if self.timeout_sec > 0:
                raise NotImplementedError

            # Zephyr actually wants a project directory, but we only get the elf now. As a workaround we
            # assume the elf is right in the build directory inside the project directory

            ret = self.platform.run(program, self)
            return ret

        def parse_stdout(self, out):
            # cpu_cycles = re.search(r"Total Cycles: (.*)", out)
            # if not cpu_cycles:
            #     logger.warning("unexpected script output (cycles)")
            #     cycles = None
            # else:
            #     cycles = int(float(cpu_cycles.group(1)))
            cpu_time_us = re.search(r"Total Time: (.*) us", out)
            if not cpu_time_us:
                logger.warning("unexpected script output (time_us)")
                time_us = None
            else:
                time_us = int(float(cpu_time_us.group(1)))
            # return cycles, time_us
            return time_us

        def get_metrics(self, elf, directory, handle_exit=None):
            if self.print_outputs:
                out = self.exec(elf, cwd=directory, live=True, handle_exit=handle_exit)
            else:
                out = self.exec(
                    elf, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=handle_exit
                )
            # cycles, time_us = self.parse_stdout(out)
            time_us = self.parse_stdout(out)

            metrics = Metrics()
            # metrics.add("Cycles", cycles)
            time_s = time_us / 1e6 if time_us is not None else time_us
            metrics.add("Runtime [s]", time_s)

            return metrics, out, []

        def get_arch(self):
            return "unkwown"

    return ZephyrPlatformTarget



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__)))

from tvm.driver.tvmc.extensions import TVMCExtension  # noqa: E402
from QVanilla_Accelerator.backend import QVanillaAcceleratorBackend  # noqa: E402


class QVanillaExtension(TVMCExtension):
    def __init__(self):
        self.backend = QVanillaAcceleratorBackend()

    def uma_backends(self):
        return [self.backend]



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility to convert various types of data into MLonMCU compatible raw binary files."""
import sys
import struct
from PIL import Image
import numpy as np


def convert(mode, val):
    """Actual convertion function.

    Arguments
    ---------
    mode : str
        The chosen mode.
    val : str
        The input data.

    Returns
    -------
    data : bytes
      Raw converted output.

    Raises
    ------
    AssertionError
        If the choses mode does not exist or value has a wrong shape.
    ValueError
        If the conversion failed.

    """
    data = b""
    assert isinstance(val, str), "Input value needs to be a string"
    assert mode in ["float", "hexstr", "int8", "bmp"], f"Unsupported mode: {mode}"

    if mode == "float":
        for f in val.split(","):
            data += struct.pack("f", float(f))
    elif mode == "hexstr":
        data = val.encode("raw_unicode_escape").decode("unicode_escape").encode("raw_unicode_escape")
    elif mode == "int8":
        for i in val.split(","):
            data += struct.pack("b", int(i))
    elif mode == "bmp":
        im = Image.open(val)
        p = np.array(im)
        assert len(p.shape) in (2, 3)  # only allow grayscale or RGB
        assert p.dtype == np.uint8  # We do not want to hande endianess at this point
        data = p.tobytes()
    return data


def write_file(dest, data):
    """Utility to save the file to disk.

    Arguments
    ---------
    dest: str
       File destination.
    data: bytes
       Raw data to export.

    """
    with open(dest, "wb") as f:
        f.write(data)


def main():
    """Main entry pint handling command line options."""
    if len(sys.argv) != 4:
        print(
            "Usage:",
            sys.argv[0],
            "mode(float, hexstr, int8, bmp)",
            "value",
            "outfile",
        )
        sys.exit(1)

    mode, val, dest = sys.argv[1], sys.argv[2], sys.argv[3]

    try:
        data = convert(mode, val)
    except ValueError:
        print("Conversion Failed")
    write_file(dest, data)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
from io import StringIO

from mlonmcu.utils import is_power_of_two, ask_user, get_base_prefix_compat, in_virtualenv


def test_utils_is_power_of_two():
    assert not is_power_of_two(0)
    assert is_power_of_two(1)
    assert is_power_of_two(2)
    assert not is_power_of_two(3)
    assert is_power_of_two(4)
    assert is_power_of_two(8)
    assert is_power_of_two(16)
    assert not is_power_of_two(99)


def test_utils_ask_user_noninteractive():
    assert not ask_user("foo", False, interactive=False)
    assert ask_user("foo", True, interactive=False)


@pytest.mark.parametrize("default", [False, True])
def test_utils_ask_user_interactive(monkeypatch, default):
    monkeypatch.setattr("sys.stdin", StringIO("y\n"))
    assert ask_user("foo", default, interactive=True)  # y
    monkeypatch.setattr("sys.stdin", StringIO("j\n"))
    assert ask_user("foo", default, interactive=True)  # j
    monkeypatch.setattr("sys.stdin", StringIO("n\n"))
    assert not ask_user("foo", default, interactive=True)  # n
    monkeypatch.setattr("sys.stdin", StringIO("Y\n"))
    assert ask_user("foo", default, interactive=True)  # Y
    monkeypatch.setattr("sys.stdin", StringIO("J\n"))
    assert ask_user("foo", default, interactive=True)  # J
    monkeypatch.setattr("sys.stdin", StringIO("N\n"))
    assert not ask_user("foo", default, interactive=True)  # N
    monkeypatch.setattr("sys.stdin", StringIO("k\n"))
    assert ask_user("foo", default, interactive=True) is default  # k
    monkeypatch.setattr("sys.stdin", StringIO("K\n"))
    assert ask_user("foo", default, interactive=True) is default  # K
    monkeypatch.setattr("sys.stdin", StringIO("\n"))
    assert ask_user("foo", default, interactive=True) is default  # empty


def test_utils_get_base_prefix_compat():
    assert get_base_prefix_compat() is not None


def test_utils_in_virtualenv():
    assert isinstance(in_virtualenv(), bool)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MicroTVM Base Platform"""
import re
import tempfile
from pathlib import Path

# from ..platform import Platform
from ..tvm.tvm_base_platform import TvmBasePlatform
from mlonmcu.setup import utils
from mlonmcu.flow.tvm.backend.python_utils import prepare_python_environment
from mlonmcu.logging import get_logger

logger = get_logger()


def parse_project_options_from_stdout(out):
    return re.compile(r"^\s+([A-Za-z0-9_]+)=", re.MULTILINE).findall(out)


def filter_project_options(valid, options):
    return {key: value for key, value in options.items() if key in valid}


def get_project_option_args(stage, project_options):
    ret = []
    for key, value in project_options.items():
        ret.append(f"{key}={value}")

    if len(ret) > 0:
        ret = ["--project-option"] + ret

    return ret


# TODO: abstarct
class MicroTvmBasePlatform(TvmBasePlatform):
    """MicroTVM base platform class."""

    DEFAULTS = {
        **TvmBasePlatform.DEFAULTS,
        "project_template": None,
        "project_options": {},
        "tvmc_custom_script": None,
        "project_dir": None,
        "experimental_tvmc_micro_tune": False,
        "experimental_tvmc_print_time": False,
    }

    REQUIRED = TvmBasePlatform.REQUIRED | {"tvm.build_dir", "tvm.pythonpath", "tvm.configs_dir"}

    def __init__(self, name, features=None, config=None):
        super().__init__(
            name,
            features=features,
            config=config,
        )
        self.tempdir = None
        self.project_name = "app"
        self.project_dir = None

    def collect_available_project_options(self, command, target=None):
        # TODO: define NotImplemented versions of the invoke_tvmc_micro_* mathods in here
        if "create" in command:
            out = self.invoke_tvmc_micro_create("_", target=target, list_options=True, live=False)
        elif command == "build":
            out = self.invoke_tvmc_micro_build(target=target, list_options=True, live=False)
        elif command == "flash":
            out = self.invoke_tvmc_micro_flash(target=target, list_options=True, live=False)
        elif command == "tune":
            tune_args = ["--output", "-", "_"]
            out = self.invoke_tvmc_micro_tune(*tune_args, target=target, list_options=True, live=False)
        elif command == "run":
            out = self.invoke_tvmc_micro_run(target=target, list_options=True, live=False)
        else:
            raise RuntimeError(f"Unexpected command: {command}")
        return parse_project_options_from_stdout(out)

    def invoke_tvmc_micro(self, command, *args, target=None, list_options=False, **kwargs):
        if list_options:
            project_option_args = ["--help"]
        else:
            options = filter_project_options(
                self.collect_available_project_options(command, target=target),
                target.get_project_options(),
            )
            project_option_args = get_project_option_args(command, options)
        return self.invoke_tvmc("micro", command, *args, *project_option_args, target=target, **kwargs)

    def get_template_args(self, target):
        template = target.template
        if target.template_path:
            template = "template"
            template_path = target.template_path
        else:
            if template == "template":
                assert self.project_template is not None
                template_path = self.project_template
            else:
                template_path = None
        if template_path:
            return ("template", "--template-dir", template_path)
        else:
            return (template,)

    def init_directory(self, path=None, context=None):
        if self.project_dir is not None:
            self.project_dir.mkdir(exist_ok=True)
            logger.debug("Project directory already initialized")
            return
        dir_name = self.name
        if path is not None:
            self.project_dir = Path(path)
        elif self.config["project_dir"] is not None:
            self.project_dir = Path(self.config["project_dir"])
        else:
            if context:
                assert "temp" in context.environment.paths
                self.project_dir = (
                    context.environment.paths["temp"].path / dir_name
                )  # TODO: Need to lock this for parallel builds
            else:
                logger.debug(
                    "Creating temporary directory because no context was available "
                    "and 'espidf.project_dir' was not supplied"
                )
                self.tempdir = tempfile.TemporaryDirectory()
                self.project_dir = Path(self.tempdir.name) / dir_name
                logger.debug("Temporary project directory: %s", self.project_dir)
        self.project_dir.mkdir(exist_ok=True)

    @property
    def project_template(self):
        return self.config["project_template"]

    @property
    def project_options(self):
        opts = self.config["project_options"]
        if isinstance(opts, str):
            opts_split = opts.split(" ")
            opts = {}
            for opt in opts_split:
                key, value = opt.split("=")[:2]
                opts[key] = value
        assert isinstance(opts, dict)
        return opts

    @property
    def tvmc_custom_script(self):
        return self.config["tvmc_custom_script"]

    @property
    def tvm_pythonpath(self):
        return self.config["tvm.pythonpath"]

    @property
    def tvm_build_dir(self):
        return self.config["tvm.build_dir"]

    @property
    def tvm_configs_dir(self):
        return self.config["tvm.configs_dir"]

    def invoke_tvmc(self, command, *args, target=None, live=None, **kwargs):
        env = prepare_python_environment(self.tvm_pythonpath, self.tvm_build_dir, self.tvm_configs_dir)
        if live is None:
            live = self.print_outputs
        if target:
            target.update_environment(env)
        if self.tvmc_custom_script is None:
            pre = ["-m", "tvm.driver.tvmc"]
        else:
            pre = [self.tvmc_custom_script]
        return utils.python(*pre, command, *args, live=live, env=env, **kwargs)

    def close(self):
        if self.tempdir:
            self.tempdir.cleanup()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of Taks Cache"""

import os
import configparser
from typing import Any


def convert_key(name):
    if not isinstance(name, tuple):
        name = (name, frozenset())
    else:
        assert len(name) == 2
        if not isinstance(name[1], frozenset):
            name = (name[0], frozenset(name[1]))
    return name


class TaskCache:
    """Task cache used to store dependency paths for the current and furture sessions.

    This can be interpreted as a "modded" dictionary which takes a key + some flags.
    """

    def __init__(self):
        self._vars = {}

    def __repr__(self):
        return str(self._vars)

    def __setitem__(self, name, value):
        name = convert_key(name)
        self._vars[name[0]] = value  # Holds latest value
        self._vars[name] = value

    def __delitem__(self, name):
        name = convert_key(name)
        del self._vars[name]

    def __getitem__(self, name):
        name = convert_key(name)
        return self._vars[name]

    def __len__(self):
        return len(self._vars)

    def __contains__(self, name):
        name = convert_key(name)
        return name in self._vars.keys()

    def find_best_match(self, name: str, flags=[]) -> Any:
        """Utility whih tries to resolve the cache entry with the beste match.

        Parameters
        ----------
        name : str
            The cache-key.
        flags : list
            Optional flags used for the lookup.
        """
        # print("find_best_match", name, flags)
        keys = self._vars.keys()
        # print("keys", keys)
        matches = []
        counts = []
        for key in keys:
            if not isinstance(key, tuple):
                continue
            assert len(key) == 2
            name_, flags_ = key[0], key[1]
            if name == name_:
                count = 0
                for flag in flags_:
                    if flag not in flags:
                        count = -1  # incompatible
                        break
                    count = count + 1
                if count >= 0:
                    matches.append(flags_)
                    counts.append(count)
        if len(counts) == 0:
            raise RuntimeError("Unable to find a match in the cache")
        m = max(counts)
        assert counts.count(m) == 1, f"For the given set of flags, there are multiple cache matches for the name {name}"
        idx = counts.index(m)
        flag = matches[idx]
        ret = self._vars[name, flag]
        return ret

    def read_from_file(self, filename, reset=True):
        if reset:
            self._vars = {}
        if not os.path.isfile(filename):
            raise RuntimeError(f"File not found: {filename}")
        cfg = configparser.ConfigParser()
        cfg.read(filename)
        sections = cfg.sections()
        for section in sections:
            if section == "default":
                flags = set()
            else:
                flags = {flag for flag in section.split(",")}
            content = dict(cfg[section].items())
            for name, value in content.items():
                self[name, flags] = value

    def write_to_file(self, filename):
        # d = self._vars

        out = {}  # This will be a dict of dicts
        for key in self._vars:
            # print(key, type(key))
            if isinstance(key, str):
                continue
            name, flags = key[0], key[1]
            value = self._vars[key]
            if len(flags) == 0:
                section_name = "default"
            else:
                section_name = ",".join(sorted(flags))
            if section_name in out:
                out[section_name][name] = value
            else:
                out[section_name] = {name: value}

        with open(filename, "w") as cachefile:
            cfg = configparser.ConfigParser()
            if "default" in out:  # Default section should be first
                cfg["default"] = out["default"]
            for x in out:
                if x == "default":
                    continue
                cfg[x] = out[x]
            cfg.write(cachefile)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Spike Target definitions"""


from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from .riscv import RISCVTarget
from .util import update_extensions

logger = get_logger()


class RVPTarget(RISCVTarget):
    """TODO"""

    FEATURES = RISCVTarget.FEATURES | {"pext"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "enable_pext": False,
        "pext_spec": 0.92,
    }
    REQUIRED = RISCVTarget.REQUIRED

    def __init__(
        self,
        name,
        features=None,
        config=None,
    ):
        super().__init__(name, features=features, config=config)
        self.supported_pext_spec_min = 0.92
        self.supported_pext_spec_max = 0.92

    @property
    def enable_pext(self):
        value = self.config["enable_pext"]
        return str2bool(value)

    @property
    def pext_spec(self):
        return float(self.config["pext_spec"])

    @property
    def extensions(self):
        exts = super().extensions
        return update_extensions(
            exts,
            pext=self.enable_pext,
            pext_spec=self.pext_spec,
        )

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        if self.enable_pext:
            major, minor = str(self.pext_spec).split(".")[:2]
            ret["RISCV_RVP_MAJOR"] = major
            ret["RISCV_RVP_MINOR"] = minor
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import json
from typing import Tuple

from .backend import TVMBackend
from mlonmcu.flow.backend import main
from mlonmcu.config import str2bool
from mlonmcu.artifact import Artifact, ArtifactFormat, lookup_artifacts
from .wrapper import generate_tvmaot_wrapper, generate_wrapper_header
from .model_info import get_relay_model_info
from .tvmc_utils import get_tvmaot_tvmc_args


class TVMAOTBackend(TVMBackend):
    FEATURES = {
        *TVMBackend.FEATURES,
        "debug_arena",
        "unpacked_api",
        "usmp",
    }

    DEFAULTS = {
        **TVMBackend.DEFAULTS,
        "debug_arena": False,
        "arena_size": None,  # Determined automatically
        "unpacked_api": False,
        "alignment_bytes": 16,
        "extra_pass_config": {
            "tir.usmp.enable": False,
        },
    }

    name = "tvmaot"

    def __init__(self, runtime="crt", fmt="mlf", system_lib=False, features=None, config=None):
        super().__init__(
            executor="aot", runtime=runtime, fmt=fmt, system_lib=system_lib, features=features, config=config
        )

    @property
    def arena_size(self):
        size = self.config["arena_size"]
        return int(size) if size is not None else None

    @property
    def unpacked_api(self):
        value = self.config["unpacked_api"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def debug_arena(self):
        value = self.config["debug_arena"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def alignment_bytes(self):
        return int(self.config["alignment_bytes"])

    def get_tvmc_compile_args(self, out, dump=None):
        return super().get_tvmc_compile_args(out, dump=dump) + get_tvmaot_tvmc_args(
            self.alignment_bytes,
            self.unpacked_api,
            system_lib=self.system_lib,
        )

    def get_workspace_size_from_metadata(self, metadata):
        if "modules" in metadata:
            return metadata["modules"]["default"]["memory"]["functions"]["main"][0]["workspace_size_bytes"]
        else:
            # backwards compatibility
            return metadata["memory"]["functions"]["main"][0]["workspace_size_bytes"]

    def generate(self) -> Tuple[dict, dict]:
        artifacts, metrics = super().generate()
        assert len(artifacts) == 1 and "default" in artifacts
        artifacts = artifacts["default"]
        assert len(metrics) == 1 and "default" in metrics
        metrics = metrics["default"]
        if self.generate_wrapper:
            if self.arena_size is not None:
                assert self.arena_size >= 0
                workspace_size = self.arena_size
            else:
                metadata_artifact = lookup_artifacts(artifacts, f"{self.prefix}.json")[0]
                metadata = json.loads(metadata_artifact.content)
                workspace_size = self.get_workspace_size_from_metadata(metadata)
            if (not self.model_info) or self.refresh_model_info:
                relay_artifact = lookup_artifacts(artifacts, f"{self.prefix}.relay")[0]
                try:
                    self.model_info = get_relay_model_info(relay_artifact.content)
                except Exception:
                    assert self.model_info is not None, "Model info missing!"
            wrapper_src = generate_tvmaot_wrapper(
                self.model_info,
                workspace_size,
                self.prefix,
                api="c" if self.unpacked_api else "packed",
                debug_arena=self.debug_arena,
            )
        artifacts.append(Artifact("aot_wrapper.c", content=wrapper_src, fmt=ArtifactFormat.SOURCE))
        header_src = generate_wrapper_header()
        artifacts.append(Artifact("tvm_wrapper.h", content=header_src, fmt=ArtifactFormat.SOURCE))
        metrics.add("Workspace Size [B]", workspace_size, True)
        return {"default": artifacts}, {"default": metrics}


if __name__ == "__main__":
    sys.exit(
        main(
            TVMAOTBackend,
            args=sys.argv[1:],
        )
    )  # pragma: no cover



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU ETISS/Pulpino Target definitions"""

import os
import re
import time
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.timeout import exec_timeout
from mlonmcu.config import str2bool
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.bench import add_bench_metrics
from .riscv import RISCVTarget

logger = get_logger()


class CV32E40PTarget(RISCVTarget):
    """Target for running CV32E40P programs in verilated RTL core."""

    FEATURES = RISCVTarget.FEATURES | {"xcorev"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "enable_xcorevmac": False,
        "enable_xcorevmem": False,
        "enable_xcorevbi": False,
        "enable_xcorevalu": False,
        "enable_xcorevbitmanip": False,
        "enable_xcorevsimd": False,
        "enable_xcorevhwlp": False,
        "fpu": "none",  # FIXED
        "atomic": False,  # FIXED
    }
    REQUIRED = RISCVTarget.REQUIRED | {"cv32e40p.verilator_executable"}

    def __init__(self, name="cv32e40p", features=None, config=None):
        super().__init__(name, features=features, config=config)
        # TODO: make optional or move to mlonmcu pkg

    @property
    def verilator_executable(self):
        return self.config["cv32e40p.verilator_executable"]

    @property
    def enable_xcorevmac(self):
        value = self.config["enable_xcorevmac"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevmem(self):
        value = self.config["enable_xcorevmem"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbi(self):
        value = self.config["enable_xcorevbi"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevalu(self):
        value = self.config["enable_xcorevalu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbitmanip(self):
        value = self.config["enable_xcorevbitmanip"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevsimd(self):
        value = self.config["enable_xcorevsimd"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevhwlp(self):
        value = self.config["enable_xcorevhwlp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def extensions(self):
        exts = super().extensions
        required = set()
        if "xcorev" not in exts:
            if self.enable_xcorevmac:
                required.add("xcvmac")
            if self.enable_xcorevmem:
                required.add("xcvmem")
            if self.enable_xcorevbi:
                required.add("xcvbi")
            if self.enable_xcorevalu:
                required.add("xcvalu")
            if self.enable_xcorevbitmanip:
                required.add("xcvbitmanip")
            if self.enable_xcorevsimd:
                required.add("xcvsimd")
            if self.enable_xcorevhwlp:
                required.add("xcvhwlp")
        for ext in required:
            if ext not in exts:
                exts.add(ext)
        return exts

    @property
    def attr(self):
        attrs = super().attr.split(",")
        if self.enable_xcorevmac:
            if "xcorevmac" not in attrs:
                attrs.append("+xcvmac")
        if self.enable_xcorevmem:
            if "xcorevmem" not in attrs:
                attrs.append("+xcvmem")
        if self.enable_xcorevbi:
            if "xcorevbi" not in attrs:
                attrs.append("+xcvbi")
        if self.enable_xcorevalu:
            if "xcorevalu" not in attrs:
                attrs.append("+xcvalu")
        if self.enable_xcorevbitmanip:
            if "xcorevbitmanip" not in attrs:
                attrs.append("+xcvbitmanip")
        if self.enable_xcorevsimd:
            if "xcorevsimd" not in attrs:
                attrs.append("+xcvsimd")
        if self.enable_xcorevhwlp:
            if "xcorevhwlp" not in attrs:
                attrs.append("+xcvhwlp")
        return ",".join(attrs)

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        # Verilator needs hex file instead of elf! Make sure it is created as an artifact!
        program_hex = Path(f"{program}.hex")
        assert program_hex.is_file(), "HEX file for CV32E40P is missing"

        verilator_executable_args = [f"+firmware={program_hex}"]
        if len(self.extra_args) > 0:
            verilator_executable_args.extend(self.extra_args.split(" "))

        if self.timeout_sec > 0:
            ret = exec_timeout(
                self.timeout_sec,
                execute,
                self.verilator_executable,
                *verilator_executable_args,
                *args,  # not really supported
                cwd=cwd,
                **kwargs,
            )
        else:
            ret = execute(
                self.verilator_executable,
                *verilator_executable_args,
                *args,  # not really supported
                cwd=cwd,
                **kwargs,
            )
        return ret

    def parse_exit(self, out):
        exit_code = super().parse_exit(out)
        if exit_code is None:
            # legacy
            exit_match = re.search(r".* EXIT FAILURE: .*", out)
            if exit_match:
                exit_code = -1  # any none-zero value
            else:
                exit_code = 0
        return exit_code

    def parse_stdout(self, out, metrics, exit_code=0):
        add_bench_metrics(out, metrics, exit_code != 0, target_name=self.name)
        sim_insns = re.search(r"#insns=(\d*)\s", out)
        if sim_insns:
            sim_insns = int(float(sim_insns.group(1)))
            metrics.add("Simulated Instructions", sim_insns, True)

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        metrics_file = os.path.join(directory, "metrics.csv")
        if os.path.exists(metrics_file):
            os.remove(metrics_file)

        # TODO: re-enable sim MIPS
        start_time = time.time()
        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
        end_time = time.time()
        diff = end_time - start_time
        exit_code = 0
        metrics = Metrics()
        self.parse_stdout(out, metrics, exit_code=exit_code)
        if metrics.has("Simulated Instructions"):
            sim_insns = metrics.get("Simulated Instructions")
            metrics.add("MIPS", (sim_insns / diff) / 1e6, True)

        return metrics, out, []

    def get_target_system(self):
        return self.name

    def get_platform_defs(self, platform):
        assert platform == "mlif"
        ret = super().get_platform_defs(platform)
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        return ret


if __name__ == "__main__":
    cli(target=CV32E40PTarget)
